# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.

import json
from typing import Optional

from authlib.integrations.starlette_client import OAuth
from fastapi import Request
from starlette.responses import RedirectResponse

from quetz.config import Config

from .base import BaseAuthenticationHandlers, BaseAuthenticator


class OAuthHandlers(BaseAuthenticationHandlers):
    """Handlers for authenticator endpoints"""

    authorize_methods = ["GET"]

    def __init__(self, authenticator, app=None):
        super().__init__(authenticator, app)

        self.router.add_api_route("/revoke", self.revoke, methods=["GET"])

    async def login(self, request: Request):
        redirect_uri = request.url_for(f'authorize_{self.authenticator.provider}')
        return await self.authenticator.client.authorize_redirect(request, redirect_uri)

    async def revoke(self, request):
        client_id = self.client.client_id
        return RedirectResponse(self.revoke_url.format(client_id=client_id))


class OAuthAuthenticator(BaseAuthenticator):
    """Base class for authenticators using Oauth2 protocol and its variants.

    The :py:meth:`authenticate` method is already implemented, but you will need to
    override some of the following variables in sublasses to make it work:

    :var str provider: name of the provider (it will be used in the url)
    :var handler_cls: class with handlers for all oauth2
      relevant endpoints in Quetz server
    :type handler_cls: sublass of :py:class:`OauthHandlers`
    :var client_id: required, client id registered with the provider
    :var client_secret: required, likewise
    :var bool is_enabled: True if authenticator is enabled, can be configured in
      :py:meth:`configure` method

    :var access_token_url: URL of the OAuth2 endpoint ot request a token
    :var authorize_url: URL of the OAuth2 ``authorize`` endpoint
    :var api_base_url: URL of the API root of the provider server
    :var validate_token_url: path of endpoint to validate the token

    """

    oauth = OAuth()
    provider = "oauth"
    handler_cls = OAuthHandlers

    # client credentials and state
    # they can be also set in configure method
    client_id: str = ""
    client_secret: str = ""
    is_enabled = False

    # oauth client params
    access_token_url: Optional[str] = None
    authorize_url: Optional[str] = None
    api_base_url: Optional[str] = None
    scope: Optional[str] = None
    server_metadata_url: Optional[str] = None
    prompt: Optional[str] = None
    client_kwargs: dict = {}

    # provider's api endpoint urls
    validate_token_url: Optional[str] = None
    revoke_url: Optional[str] = None

    @property
    def router(self):
        return self.handler.router

    def __init__(self, config: Config, client_kwargs=None, provider=None, app=None):
        super().__init__(config, provider, app)
        if self.is_enabled:
            self.register(client_kwargs=client_kwargs)

    async def userinfo(self, request, token):
        raise NotImplementedError("subclasses need to implement userinfo")

    async def authenticate(self, request, data=None, dao=None, config=None):
        token = await self.client.authorize_access_token(request)

        profile = await self.userinfo(request, token)

        username = profile["login"]
        auth_state = {"token": json.dumps(token), "provider": self.provider}

        return {"username": username, "profile": profile, "auth_state": auth_state}

    def register(self, client_kwargs=None):
        if client_kwargs is None:
            client_kwargs = {}

        if self.scope:
            client_kwargs["scope"] = self.scope

        client_kwargs.update(self.client_kwargs)

        self.client = self.oauth.register(
            name=self.provider,
            client_id=self.client_id,
            client_secret=self.client_secret,
            access_token_url=self.access_token_url,
            authorize_url=self.authorize_url,
            api_base_url=self.api_base_url,
            server_metadata_url=self.server_metadata_url,
            prompt=self.prompt,
            client_kwargs=client_kwargs,
        )

    async def validate_token(self, token):
        try:
            resp = await self.client.get(
                self.validate_token_url, token=json.loads(token)
            )
            return resp.status_code != 401
        except Exception:
            return False



"""add per channel TTL

Revision ID: cd404ed93cc0
Revises: 303ff70c27fc
Create Date: 2021-01-07 20:12:23.527311

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'cd404ed93cc0'
down_revision = '303ff70c27fc'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'channels',
        sa.Column('ttl', sa.Integer(), server_default='36000', nullable=False),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('channels', 'ttl')
    # ### end Alembic commands ###



import json
import signal
from multiprocessing import active_children

from starlette.requests import Request as ASGIRequest
from starlette.responses import Response as ASGIResponse


class AsyncPathMapDispatch:
    # dummy server, copied from authlib tests
    def __init__(self, path_maps):
        self.path_maps = path_maps

    async def __call__(self, scope, receive, send):
        request = ASGIRequest(scope, receive=receive)

        rv = self.path_maps[request.url.path]
        status_code = rv.get('status_code', 200)
        body = rv.get('body')
        headers = rv.get('headers', {})
        if isinstance(body, dict):
            body = json.dumps(body).encode()
            headers['Content-Type'] = 'application/json'
        else:
            if isinstance(body, str):
                body = body.encode()
            headers['Content-Type'] = 'application/x-www-form-urlencoded'

        response = ASGIResponse(
            status_code=status_code,
            content=body,
            headers=headers,
        )
        await response(scope, receive, send)


class Interrupt:
    # Interrupt child when SIGALRM is received.
    # Useful to kill the server when it is correctly launched, using a timeout.
    def _handle_interrupt(self, signum, frame):
        for p in active_children():
            p.terminate()
            p.join()

    def __enter__(self):
        signal.signal(signal.SIGALRM, self._handle_interrupt)

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



# Copyright 2020 QuantStack, Codethink Ltd
# Distributed under the terms of the Modified BSD License.

from quetz.config import Config

from .oauth2 import OAuthAuthenticator


class GoogleAuthenticator(OAuthAuthenticator):
    """Use Google account to authenticate users with Quetz.

    To enable add the following to the configuration file:

    .. code::

      [google]
      client_id = "1111111111-dha39auqzp92110sdf.apps.googleusercontent.com"
      client_secret = "03728444a12abff17e9444fd231b4379d58f0b"

    You can obtain ``client_id`` and ``client_secret`` by registering your
    application with Google platfrom at this URL:
    `<https://console.developers.google.com/apis/credentials>`_.
    """

    provider = 'google'
    server_metadata_url = 'https://accounts.google.com/.well-known/openid-configuration'
    scope = "openid email profile"
    prompt = 'select_account'

    revoke_url = 'https://myaccount.google.com/permissions'
    validate_token_url = 'https://openidconnect.googleapis.com/v1/userinfo'

    collect_emails = False

    async def userinfo(self, request, token):
        profile = await self.client.parse_id_token(request, token)

        github_profile = {
            "id": profile["sub"],
            "name": profile["name"],
            "avatar_url": profile['picture'],
            "email": profile["email"],
            "login": profile["email"],
        }

        if self.collect_emails:
            github_profile["emails"] = [
                {
                    "email": profile["email"],
                    "primary": True,
                    "verified": profile["email_verified"],
                }
            ]

        return github_profile

    def configure(self, config: Config):
        if config.configured_section("google"):
            self.client_id = config.google_client_id
            self.client_secret = config.google_client_secret
            self.is_enabled = True
            if config.configured_section("users"):
                self.collect_emails = config.users_collect_emails

        else:
            self.is_enabled = False

        # call the configure of base class to set default_channel and default role
        super().configure(config)



import os
import pickle
import sys
import time
import uuid
from datetime import datetime, timedelta
from pathlib import Path

import pytest

from quetz.config import Config
from quetz.dao import Dao
from quetz.db_models import User
from quetz.jobs.dao import JobsDao
from quetz.jobs.models import Job, JobStatus, Task, TaskStatus
from quetz.jobs.runner import Supervisor, mk_sql_expr, parse_conda_spec
from quetz.rest_models import Channel, Package
from quetz.tasks.workers import SubprocessWorker
from quetz.testing.mockups import MockWorker

pytest_plugins = ("pytest_asyncio",)


@pytest.fixture
def auto_rollback():
    return False


@pytest.fixture
def sqlite_in_memory():
    return False


def add_package_version(
    filename, package_version, channel_name, user, dao, package_name=None
):
    if not package_name:
        package_name = "test-package"

    package_format = "tarbz2"
    package_info = "{}"
    path = Path(filename)
    version = dao.create_version(
        channel_name,
        package_name,
        package_format,
        "linux-64",
        package_version,
        0,
        "",
        path.name,
        package_info,
        user.id,
        size=11,
    )
    return version


@pytest.fixture
def package_version(
    db,
    user,
    public_channel,
    channel_name,
    package_name,
    public_package,
    dao: Dao,
    config: Config,
):
    pkgstore = config.get_package_store()

    filename = "test-package-0.1-0.tar.bz2"
    version = add_package_version(
        filename, "0.1", channel_name, user, dao, package_name
    )
    path = Path(filename)
    with open(path, "rb") as fid:
        pkgstore.add_file(fid.read(), channel_name, "linux-64" / path)

    dao.update_channel_size(channel_name)
    db.refresh(public_channel)

    yield version

    db.delete(version)
    db.commit()


@pytest.fixture
def public_channel(dao: Dao, user, channel_role, channel_name, db):
    channel_data = Channel(name=channel_name, private=False)
    channel = dao.create_channel(channel_data, user.id, channel_role)

    yield channel

    db.delete(channel)
    db.commit()


@pytest.fixture
def public_package(db, user, public_channel, dao, package_role, package_name):
    package_data = Package(name=package_name)

    package = dao.create_package(
        public_channel.name, package_data, user.id, package_role
    )

    return package


@pytest.fixture
def manager(config, db):
    manager = SubprocessWorker(config)
    yield manager
    manager._executor.shutdown()
    SubprocessWorker._executor = None
    db.query(Job).delete()
    db.commit()


@pytest.fixture
def supervisor(db, config, manager):
    supervisor = Supervisor(db, manager)
    return supervisor


def test_create_task(db, user, package_version):
    job = Job(owner_id=user.id, manifest=b"")
    task = Task(job=job)
    db.add(job)
    db.add(task)
    db.commit()


def func(package_version: dict, config: Config):
    with open("test-output.txt", "a") as fid:
        fid.write("ok")


def failed_func(package_version: dict):
    raise Exception("some exception")


def long_running(package_version: dict):
    time.sleep(0.25)


def dummy_func(package_version: dict):
    pass


@pytest.mark.asyncio
async def test_create_job(db, user, package_version, supervisor):
    func_serialized = pickle.dumps(func)
    job = Job(owner_id=user.id, manifest=func_serialized, items_spec="*")
    db.add(job)
    db.commit()
    supervisor.run_jobs()
    new_jobs = supervisor.run_tasks()
    db.refresh(job)
    task = db.query(Task).one()

    assert job.status == JobStatus.running

    assert task.status == TaskStatus.pending

    # wait for job to finish
    await new_jobs[0].wait()

    supervisor.check_status()

    db.refresh(task)
    assert task.status == TaskStatus.success
    assert os.path.isfile("test-output.txt")

    db.refresh(job)
    assert job.status == JobStatus.success


@pytest.mark.asyncio
async def test_run_tasks_only_on_new_versions(
    db, user, package_version, dao, channel_name, package_name, supervisor
):
    func_serialized = pickle.dumps(dummy_func)
    job = Job(owner_id=user.id, manifest=func_serialized, items_spec="*")
    db.add(job)
    db.commit()
    supervisor.run_jobs()
    new_jobs = supervisor.run_tasks()
    db.refresh(job)
    task = db.query(Task).one()

    await new_jobs[0].wait()
    supervisor.check_status()
    db.refresh(task)
    db.refresh(job)
    assert task.status == TaskStatus.success
    assert job.status == JobStatus.success

    job.status = JobStatus.pending
    db.commit()
    supervisor.run_jobs()
    new_jobs = supervisor.run_tasks()
    supervisor.check_status()
    db.refresh(job)
    assert not new_jobs
    assert job.status == JobStatus.success

    filename = "test-package-0.2-0.tar.bz2"
    add_package_version(filename, "0.2", channel_name, user, dao, package_name)

    job.status = JobStatus.pending
    db.commit()
    supervisor.run_jobs()
    new_jobs = supervisor.run_tasks()
    assert len(new_jobs) == 1
    assert job.status == JobStatus.running
    assert len(job.tasks) == 2
    assert job.tasks[0].status == TaskStatus.success
    assert job.tasks[1].status == TaskStatus.pending

    # force rerunning
    job.status = JobStatus.pending
    supervisor.run_jobs(force=True)
    db.refresh(job)
    new_jobs = supervisor.run_tasks()
    assert len(job.tasks) == 4
    assert len(new_jobs) == 2


@pytest.mark.asyncio
async def test_running_task(db, user, package_version, supervisor):
    func_serialized = pickle.dumps(long_running)
    job = Job(owner_id=user.id, manifest=func_serialized, items_spec="*")
    db.add(job)
    db.commit()
    supervisor.run_jobs()
    processes = supervisor.run_tasks()
    db.refresh(job)
    task = db.query(Task).one()

    assert job.status == JobStatus.running

    assert task.status == TaskStatus.pending

    # wait for task status to change
    for i in range(50):
        time.sleep(0.05)

        db.refresh(task)
        if task.status != TaskStatus.pending:
            break

    assert task.status == TaskStatus.running

    # wait for job to finish
    await processes[0].wait()

    supervisor.check_status()

    db.refresh(task)
    assert task.status == TaskStatus.success


@pytest.mark.parametrize("items_spec", [None, "*"])
@pytest.mark.asyncio
async def test_restart_worker_process(
    db, user, package_version, supervisor, caplog, items_spec
):
    # test if we can resume jobs if a worker was killed/restarted
    func_serialized = pickle.dumps(long_running)

    job = Job(owner_id=user.id, manifest=func_serialized, items_spec="*")
    db.add(job)
    db.commit()
    supervisor.run_jobs()
    supervisor.run_tasks()
    db.refresh(job)
    task = db.query(Task).one()

    assert job.status == JobStatus.running

    assert task.status == TaskStatus.pending

    # wait for task status to change
    for i in range(50):
        time.sleep(0.05)

        db.refresh(task)
        if task.status != TaskStatus.pending:
            break

    db.refresh(task)
    assert task.status == TaskStatus.running

    # simulate restart
    supervisor = Supervisor(db, supervisor.manager)

    db.refresh(job)
    assert job.status == JobStatus.running
    assert task.status == TaskStatus.failed
    assert "failed" in caplog.text

    supervisor.run_once()

    db.refresh(job)

    assert job.status == JobStatus.failed


@pytest.mark.asyncio
async def test_failed_task(db, user, package_version, supervisor):
    func_serialized = pickle.dumps(failed_func)
    job = Job(owner_id=user.id, manifest=func_serialized, items_spec="*")
    db.add(job)
    db.commit()
    supervisor.run_jobs()
    new_jobs = supervisor.run_tasks()
    task = db.query(Task).one()
    with pytest.raises(Exception, match="some exception"):
        await new_jobs[0].wait()

    supervisor.check_status()

    db.refresh(task)
    assert task.status == TaskStatus.failed

    db.refresh(job)
    assert job.status == JobStatus.failed


@pytest.mark.parametrize("items_spec", [""])
def test_empty_package_spec(db, user, package_version, caplog, items_spec, supervisor):
    func_serialized = pickle.dumps(func)
    job = Job(owner_id=user.id, manifest=func_serialized, items_spec=items_spec)
    db.add(job)
    db.commit()
    supervisor.run_jobs()
    db.refresh(job)
    task = db.query(Task).one_or_none()

    assert job.status == JobStatus.success
    assert task is None


def test_mk_query():
    def compile(dict_spec):
        s = mk_sql_expr(dict_spec)
        sql_expr = str(s.compile(compile_kwargs={"literal_binds": True}))
        return sql_expr

    spec = [{"version": ("eq", "0.1"), "package_name": ("in", ["my-package"])}]
    sql_expr = compile(spec)

    assert sql_expr == (
        "package_versions.version = '0.1' "
        "AND package_versions.package_name IN ('my-package')"
    )

    spec = [{"version": ("lt", "0.2")}]
    sql_expr = compile(spec)

    assert sql_expr == "package_versions.version < '0.2'"

    spec = [{"version": ("lte", "0.2")}]
    sql_expr = compile(spec)

    assert sql_expr == "package_versions.version <= '0.2'"

    spec = [{"version": ("gte", "0.3")}]
    sql_expr = compile(spec)

    assert sql_expr == "package_versions.version >= '0.3'"

    spec = [
        {"version": ("lt", "0.2"), "package_name": ("eq", "my-package")},
        {"version": ("gt", "0.3"), "package_name": ("eq", "other-package")},
    ]
    sql_expr = compile(spec)

    assert sql_expr == (
        "package_versions.version < '0.2' "
        "AND package_versions.package_name = 'my-package' "
        "OR package_versions.version > '0.3' "
        "AND package_versions.package_name = 'other-package'"
    )

    spec = [
        {"version": ("and", ("lt", "0.2"), ("gt", "0.1"))},
    ]
    sql_expr = compile(spec)

    assert sql_expr == (
        "package_versions.version < '0.2'" " AND package_versions.version > '0.1'"
    )

    spec = [
        {"version": ("or", ("and", ("lt", "0.2"), ("gt", "0.1")), ("gt", "0.3"))},
    ]
    sql_expr = compile(spec)
    assert sql_expr == (
        "package_versions.version < '0.2'"
        " AND package_versions.version > '0.1'"
        " OR package_versions.version > '0.3'"
    )

    spec = [{"package_name": ("like", "my-*")}]
    sql_expr = compile(spec)

    assert sql_expr == ("lower(package_versions.package_name) LIKE lower('my-%')")

    spec = [{"package_name": ("like", "*")}]
    sql_expr = compile(spec)

    assert sql_expr == ("lower(package_versions.package_name) LIKE lower('%')")


def test_parse_conda_spec():
    dict_spec = parse_conda_spec("*")
    assert dict_spec == [{"package_name": ("like", "*")}]

    dict_spec = parse_conda_spec("my-package==0.1.1")
    assert dict_spec == [
        {"version": ("eq", "0.1.1"), "package_name": ("eq", "my-package")}
    ]

    dict_spec = parse_conda_spec("my-package==0.1.2,other-package==0.5.1")
    assert dict_spec == [
        {"version": ("eq", "0.1.2"), "package_name": ("eq", "my-package")},
        {"version": ("eq", "0.5.1"), "package_name": ("eq", "other-package")},
    ]
    dict_spec = parse_conda_spec("my-package>0.1.2")
    assert dict_spec == [
        {"version": ("gt", "0.1.2"), "package_name": ("eq", "my-package")},
    ]

    dict_spec = parse_conda_spec("my-package<0.1.2")
    assert dict_spec == [
        {"version": ("lt", "0.1.2"), "package_name": ("eq", "my-package")},
    ]

    dict_spec = parse_conda_spec("my-package>=0.1.2")
    assert dict_spec == [
        {"version": ("gte", "0.1.2"), "package_name": ("eq", "my-package")},
    ]

    dict_spec = parse_conda_spec("my-package-v2==0.1")
    assert dict_spec == [
        {"version": ("eq", "0.1"), "package_name": ("eq", "my-package-v2")},
    ]

    dict_spec = parse_conda_spec("my-package>=0.1.2,<0.2")
    assert dict_spec == [
        {
            "version": ("and", ("gte", "0.1.2"), ("lt", "0.2")),
            "package_name": ("eq", "my-package"),
        },
    ]

    dict_spec = parse_conda_spec("my-package>=0.1.2,<0.2,other-package==1.1")
    assert dict_spec == [
        {
            "version": ("and", ("gte", "0.1.2"), ("lt", "0.2")),
            "package_name": ("eq", "my-package"),
        },
        {
            "version": ("eq", "1.1"),
            "package_name": ("eq", "other-package"),
        },
    ]

    dict_spec = parse_conda_spec("my-package")
    assert dict_spec == [{"package_name": ("eq", "my-package")}]

    dict_spec = parse_conda_spec("my-*")
    assert dict_spec == [{"package_name": ("like", "my-*")}]


@pytest.mark.parametrize(
    "spec,n_tasks",
    [
        ("my-package==0.1", 1),
        ("my-package==0.2", 0),
        ("my-package==0.1,my-package==0.2", 1),
        ("my-package", 1),
        ("*", 1),
    ],
)
def test_filter_versions(db, user, package_version, spec, n_tasks, supervisor):
    func_serialized = pickle.dumps(func)
    job = Job(
        owner_id=user.id,
        manifest=func_serialized,
        items_spec=spec,
    )
    db.add(job)
    db.commit()
    supervisor.run_jobs()
    supervisor.run_tasks()
    db.refresh(job)
    n_created_tasks = db.query(Task).count()

    assert n_created_tasks == n_tasks


@pytest.mark.parametrize("user_role", ["owner"])
def test_refresh_job(auth_client, user, db, package_version, supervisor):
    func_serialized = pickle.dumps(dummy_func)
    job = Job(
        owner_id=user.id,
        manifest=func_serialized,
        items_spec="*",
        status=JobStatus.success,
    )
    task = Task(job=job, status=TaskStatus.success, package_version=package_version)
    db.add(job)
    db.add(task)
    db.commit()

    assert job.status == JobStatus.success
    assert len(job.tasks) == 1

    response = auth_client.patch(f"/api/jobs/{job.id}", json={"status": "pending"})

    assert response.status_code == 200

    db.refresh(job)
    assert job.status == JobStatus.pending
    assert len(job.tasks) == 1

    supervisor.run_jobs()
    supervisor.check_status()
    assert job.status == JobStatus.success
    assert len(job.tasks) == 1

    response = auth_client.patch(
        f"/api/jobs/{job.id}", json={"status": "pending", "force": True}
    )
    supervisor.run_jobs()
    db.refresh(job)
    assert job.status == JobStatus.running
    assert len(job.tasks) == 2

    # forcing one job should not affect the other
    other_job = Job(
        id=2,
        owner_id=user.id,
        manifest=func_serialized,
        items_spec="*",
        status=JobStatus.success,
    )
    job.status = JobStatus.pending
    db.add(other_job)
    db.commit()

    response = auth_client.patch(
        f"/api/jobs/{other_job.id}", json={"status": "pending", "force": True}
    )
    db.refresh(job)
    assert job.status == JobStatus.pending
    assert len(job.tasks) == 2


@pytest.fixture(scope="session")
def dummy_job_plugin(test_data_dir):
    plugin_dir = str(Path(test_data_dir) / "dummy-plugin")
    sys.path.insert(0, plugin_dir)

    yield

    sys.path.remove(plugin_dir)


@pytest.mark.parametrize("user_role", ["owner"])
@pytest.mark.parametrize(
    "manifest",
    [
        "dummy_func",
        "os:listdir",
        "os.listdir",
        "quetz.dummy_func",
        "quetz-plugin:dummy_func",
        "quetz-dummyplugin:missing_job",
        "quetz-dummyplugin:dummy_job:error",
    ],
)
def test_post_new_job_manifest_validation(
    auth_client, user, db, manifest, dummy_job_plugin
):
    response = auth_client.post(
        "/api/jobs", json={"items_spec": "*", "manifest": manifest}
    )
    assert response.status_code == 422
    msg = response.json()['detail'][0]['msg']
    assert "invalid function" in msg
    for name in manifest.split(":"):
        assert name in msg


@pytest.mark.parametrize("user_role", ["owner"])
def test_post_new_job_invalid_items_spec(auth_client, user, db, dummy_job_plugin):
    # items_spec=None is not allowed for jobs
    # (but it works with actions)
    manifest = "quetz-dummyplugin:dummy_func"
    response = auth_client.post(
        "/api/jobs", json={"items_spec": None, "manifest": manifest}
    )
    assert response.status_code == 422
    msg = response.json()['detail']
    assert "not an allowed value" in msg[0]['msg']


@pytest.mark.parametrize("user_role", ["owner"])
@pytest.mark.parametrize(
    "manifest", ["quetz-dummyplugin:dummy_func", "quetz-dummyplugin:dummy_job"]
)
def test_post_new_job_from_plugin(
    auth_client,
    user,
    db,
    manifest,
    dummy_job_plugin,
    sync_supervisor,
    package_version,
    mocker,
):
    dummy_func = mocker.Mock()
    mocker.patch("quetz_dummyplugin.jobs.dummy_func", dummy_func, create=True)
    response = auth_client.post(
        "/api/jobs", json={"items_spec": "*", "manifest": manifest}
    )
    assert response.status_code == 201
    job_id = response.json()['id']
    job = db.get(Job, job_id)
    assert job.manifest.decode('ascii') == manifest

    sync_supervisor.run_once()

    assert job.status == JobStatus.success
    assert job.tasks

    if manifest.endswith("dummy_func"):
        dummy_func.assert_called_once()
    else:
        dummy_func.assert_not_called()


@pytest.mark.parametrize("user_role", ["owner"])
def test_post_new_job_with_handler(
    auth_client, user, db, mock_action, sync_supervisor, package_version
):
    response = auth_client.post(
        "/api/jobs", json={"items_spec": "*", "manifest": "test_action"}
    )
    assert response.status_code == 201
    job_id = response.json()['id']
    job = db.get(Job, job_id)
    assert job.status == JobStatus.pending
    assert job.manifest.decode('ascii') == "test_action"

    sync_supervisor.run_once()

    assert job.status == JobStatus.success
    assert job.tasks

    mock_action.assert_called_once()


@pytest.mark.parametrize("user_role", ["owner"])
@pytest.mark.parametrize(
    "status,job_ids",
    [
        (["pending"], [1]),
        (["running"], [0]),
        (["pending", "running"], [0, 1]),
        ([], [0, 1]),
        (["success"], []),
    ],
)
def test_filter_jobs_by_status(auth_client, db, user, status, job_ids):
    job0 = Job(
        id=0,
        items_spec="*",
        manifest=b"dummy_func",
        status=JobStatus.running,
        owner=user,
    )
    db.add(job0)
    job1 = Job(
        id=1,
        items_spec="*",
        manifest=b"dummy_func",
        status=JobStatus.pending,
        owner=user,
    )
    db.add(job1)

    db.commit()

    query = "&".join([f"status={s}" for s in status])

    response = auth_client.get(f"/api/jobs?{query}")

    assert response.status_code == 200
    response_data = response.json()
    assert {job['id'] for job in response_data['result']} == set(job_ids)
    assert response_data['pagination']['all_records_count'] == len(job_ids)


@pytest.mark.parametrize("user_role", ["owner"])
@pytest.mark.parametrize(
    "query_str,ok",
    [
        ("status=pending&status=running", True),
        ("status=missing", False),
        ("status=success", True),
        ("status=failed", True),
        ("status=pending,running", False),
        ("limit=wrongtype", False),
        ("limit=10", True),
    ],
)
def test_validate_query_string(auth_client, query_str, ok):
    response = auth_client.get(f"/api/jobs?{query_str}")
    if ok:
        assert response.status_code == 200
    else:
        assert response.status_code == 422


@pytest.fixture
def many_jobs(db, user):
    n_jobs = 5
    jobs = []
    for i in range(n_jobs):
        job = Job(
            id=i,
            items_spec="*",
            manifest=b"dummy_func",
            status=JobStatus.running,
            owner=user,
        )
        db.add(job)
        jobs.append(job)

    yield jobs

    for job in jobs:
        db.delete(job)
    db.commit()


@pytest.mark.parametrize("user_role", ["owner"])
@pytest.mark.parametrize("skip,limit", [(0, 2), (1, -1), (2, 1), (2, 5)])
def test_jobs_pagination(auth_client, skip, limit, many_jobs):
    n_jobs = len(many_jobs)

    response = auth_client.get(f"/api/jobs?skip={skip}&limit={limit}")
    assert response.status_code == 200

    jobs = response.json()["result"]

    assert jobs[0]['id'] == skip
    if limit > 0:
        assert jobs[-1]['id'] == min(n_jobs - 1, skip + limit - 1)
        assert len(jobs) == min(n_jobs - skip, limit)
    else:
        assert jobs[-1]['id'] == n_jobs - 1


@pytest.mark.parametrize(
    "query_params,expected_task_count",
    [
        ("", 1),
        ("status=created&status=pending", 1),
        ("skip=0&limit=1", 1),
        ("skip=1", 0),
        ("limit=0", 0),
        ("status=running", 0),
    ],
)
@pytest.mark.parametrize("user_role", ["owner"])
def test_get_tasks(
    auth_client, db, user, package_version, query_params, expected_task_count
):
    job = Job(items_spec="*", owner=user, manifest=pickle.dumps(dummy_func))
    db.add(job)
    task = Task(job=job, package_version=package_version)
    db.add(task)

    db.commit()

    response = auth_client.get(f"/api/jobs/{job.id}/tasks?{query_params}")
    assert response.status_code == 200
    data = response.json()

    assert len(data['result']) == expected_task_count

    if expected_task_count > 0:
        assert data['result'][0]['id'] == task.id
        assert data['result'][0]['job_id'] == job.id


@pytest.fixture()
def other_user(db):
    other_user = User(id=uuid.uuid4().bytes, username='otheruser')

    db.add(other_user)

    yield other_user

    db.delete(other_user)
    db.commit()


@pytest.mark.parametrize("user_role", ["member"])
def test_get_user_jobs(auth_client, db, user, package_version, other_user):
    job = Job(items_spec="*", owner=user, manifest=b"dummy_func")
    db.add(job)

    other_job = Job(items_spec="*", owner=other_user, manifest=b"dummy_func")
    db.add(other_job)

    db.commit()

    response = auth_client.get("/api/jobs")
    assert response.status_code == 200
    data = response.json()['result']
    assert len(data) == 1


@pytest.fixture
def action_job(db, user):
    job_dao = JobsDao(db)
    job = job_dao.create_job(b"test_action", user.id, extra_args={"my_arg": 1})

    yield job

    db.delete(job)
    db.commit()


@pytest.fixture
def package_version_job(db, user, package_version):
    func_serialized = pickle.dumps(dummy_func)
    job = Job(owner=user, manifest=func_serialized, items_spec="*")
    db.add(job)
    db.commit()
    yield job

    db.delete(job)
    db.commit()


@pytest.fixture
def sync_supervisor(db, dao, config):
    "supervisor with synchronous test worker"
    manager = MockWorker(config, db, dao)
    supervisor = Supervisor(db, manager)
    return supervisor


@pytest.fixture
def mock_action(mocker):
    func = mocker.Mock()
    mocker.patch("quetz.jobs.handlers.JOB_HANDLERS", {"test_action": func})
    return func


def test_update_job_status(sync_supervisor, db, action_job):
    running_task = Task(status=TaskStatus.running)
    finished_task = Task(status=TaskStatus.success)
    action_job.tasks.append(running_task)
    action_job.tasks.append(finished_task)
    action_job.status = JobStatus.running

    db.commit()

    sync_supervisor.run_once()

    db.refresh(action_job)
    db.refresh(running_task)
    db.refresh(finished_task)

    assert finished_task.status == TaskStatus.success
    assert running_task.status == TaskStatus.running
    assert action_job.status == JobStatus.running


def test_run_action_handler(sync_supervisor, db, caplog, action_job, mock_action):
    sync_supervisor.run_once()
    assert "ERROR" not in caplog.text
    mock_action.assert_called_with(my_arg=1)
    db.refresh(action_job)
    assert action_job.tasks[0].status == TaskStatus.success
    assert action_job.status == JobStatus.success


def test_update_periodic_action(sync_supervisor, db, action_job, mock_action):
    job = action_job
    job.repeat_every_seconds = 10
    db.commit()

    sync_supervisor.run_once()
    db.refresh(action_job)
    assert job.tasks[0].status == TaskStatus.success
    assert job.status == JobStatus.pending


@pytest.mark.parametrize("start_date", [datetime(1960, 1, 1, 10, 0, 0), None])
def test_run_action_once(sync_supervisor, db, action_job, mock_action, start_date):
    # job should start immediatedly
    action_job.start_at = start_date
    db.commit()

    assert action_job.status == JobStatus.pending

    sync_supervisor.run_once()

    db.refresh(action_job)

    assert len(action_job.tasks) == 1
    assert action_job.status == JobStatus.success

    sync_supervisor.run_once()

    db.refresh(action_job)
    assert len(action_job.tasks) == 1
    assert action_job.status == JobStatus.success


def test_run_action_after_delay(sync_supervisor, db, action_job, mock_action, mocker):
    action_job.start_at = datetime(3020, 1, 1, 10, 0)
    db.commit()

    assert action_job.status == JobStatus.pending

    sync_supervisor.run_once()

    db.refresh(action_job)

    assert len(action_job.tasks) == 0
    assert action_job.status == JobStatus.pending

    sync_supervisor.run_once()

    mock_datetime = mocker.patch("quetz.jobs.runner.datetime")
    mock_datetime.utcnow.return_value = datetime(3020, 1, 1, 10, 1)

    sync_supervisor.run_once()
    assert len(action_job.tasks) == 1
    assert action_job.status == JobStatus.success


@pytest.mark.parametrize("start_date", [datetime(1960, 1, 1, 10, 0, 0), None])
def test_run_periodic_action(
    sync_supervisor, db, action_job, mock_action, mocker, start_date
):
    action_job.repeat_every_seconds = 10
    action_job.start_date = start_date
    now = datetime.utcnow()
    delta = timedelta(seconds=11)

    db.commit()

    sync_supervisor.run_once()

    assert len(action_job.tasks) == 1
    assert action_job.status == JobStatus.pending

    sync_supervisor.run_once()

    assert len(action_job.tasks) == 1
    assert action_job.status == JobStatus.pending

    mock_datetime = mocker.patch("quetz.jobs.runner.datetime")
    mock_datetime.utcnow.return_value = now + delta

    sync_supervisor.run_once()
    assert len(action_job.tasks) == 2
    assert action_job.status == JobStatus.pending

    sync_supervisor.run_once()
    assert len(action_job.tasks) == 2
    assert action_job.status == JobStatus.pending


@pytest.mark.parametrize("start_date", [datetime(1960, 1, 1, 10, 0, 0), None])
def test_run_periodic_package_version_job(
    sync_supervisor, db, package_version_job, mocker, start_date
):
    package_version_job.repeat_every_seconds = 10
    package_version_job.start_date = start_date

    now = datetime.utcnow()
    delta = timedelta(seconds=11)

    db.commit()

    sync_supervisor.run_once()

    assert len(package_version_job.tasks) == 1
    assert package_version_job.tasks[0].status == TaskStatus.success
    assert package_version_job.status == JobStatus.pending

    sync_supervisor.run_once()
    assert len(package_version_job.tasks) == 1
    assert package_version_job.tasks[0].status == TaskStatus.success
    assert package_version_job.status == JobStatus.pending

    mock_datetime = mocker.patch("quetz.jobs.runner.datetime")
    mock_datetime.utcnow.return_value = now + delta

    sync_supervisor.run_once()
    assert len(package_version_job.tasks) == 2
    assert package_version_job.status == JobStatus.pending

    sync_supervisor.run_once()
    assert len(package_version_job.tasks) == 2
    assert package_version_job.status == JobStatus.pending



import contextlib
import json
import logging
import os
import shutil
from concurrent.futures import ThreadPoolExecutor
from http.client import IncompleteRead
from tempfile import SpooledTemporaryFile
from typing import List

import requests
from fastapi import HTTPException, status
from tenacity import TryAgain, after_log, retry, stop_after_attempt, wait_exponential

from quetz import authorization, rest_models
from quetz.condainfo import CondaInfo, get_subdir_compat
from quetz.config import Config
from quetz.dao import Dao
from quetz.db_models import PackageVersion
from quetz.errors import DBError
from quetz.pkgstores import PackageStore
from quetz.tasks import indexing
from quetz.utils import TicToc, add_static_file, check_package_membership

# copy common subdirs from conda:
# https://github.com/conda/conda/blob/a78a2387f26a188991d771967fc33aa1fb5bb810/conda/base/constants.py#L63

KNOWN_SUBDIRS = (
    "noarch",
    "linux-32",
    "linux-64",
    "linux-aarch64",
    "linux-armv6l",
    "linux-armv7l",
    "linux-ppc64",
    "linux-ppc64le",
    "linux-s390x",
    "osx-64",
    "osx-arm64",
    "win-32",
    "win-64",
    "zos-z",
)


logger = logging.getLogger("quetz")


class RemoteRepository:
    """Resource object for external package repositories."""

    def __init__(self, host, session):
        self.host = host
        self.session = session

    def open(self, path):
        return RemoteFile(self.host, path, self.session)


class RemoteServerError(Exception):
    pass


class RemoteFileNotFound(RemoteServerError):
    pass


class RemoteFile:
    def __init__(self, host: str, path: str, session=None):
        if session is None:
            session_as_param = False
            session = requests.Session()
        else:
            session_as_param = True
        remote_url = os.path.join(host, path)
        try:
            response = session.get(remote_url, stream=True)
        except requests.ConnectionError:
            raise RemoteServerError
        if response.status_code == 404:
            raise RemoteFileNotFound
        elif response.status_code != 200:
            raise RemoteServerError
        self.file = SpooledTemporaryFile()
        response.raw.decode_content = True  # for gzipped response content
        shutil.copyfileobj(response.raw, self.file)
        response.close()
        if not session_as_param:
            session.close()

        # workaround for https://github.com/python/cpython/pull/3249
        if not hasattr(self.file, "seekable"):
            self.file.readable = self.file._file.readable  # type: ignore
            self.file.writable = self.file._file.writable  # type: ignore
            self.file.seekable = self.file._file.seekable  # type: ignore

        # rewind
        self.file.seek(0)
        _, self.filename = os.path.split(remote_url)
        self.content_type = response.headers.get("content-type")

    def json(self):
        return json.load(self.file)


def download_remote_file(
    repository: RemoteRepository, pkgstore: PackageStore, channel: str, path: str
):
    """Download a file from a remote repository to a package store"""

    # Check if a download is already underway for this file
    lock = pkgstore.get_download_lock(channel, path)
    if lock:
        # Wait for the download to complete
        lock.acquire()
        # Release the lock so that any other clients can also finish
        lock.release()
        return
    # Acquire a lock to prevent multiple concurrent downloads of the same file
    with pkgstore.create_download_lock(channel, path):
        logger.debug(f"Downloading {path} from {channel} to pkgstore")
        remote_file = repository.open(path)
        data_stream = remote_file.file

        if path.endswith('.json'):
            add_static_file(data_stream.read(), channel, None, path, pkgstore)
        else:
            pkgstore.add_package(data_stream, channel, path)

    pkgstore.delete_download_lock(channel, path)


@contextlib.contextmanager
def _check_checksum(dao: Dao, channel_name: str, platform: str, keyname="sha256"):
    """context manager to compare sha or md5 hashes"""

    # lazily load repodata on first request
    package_fingerprints = None

    def _func(package_name, metadata):
        nonlocal package_fingerprints

        if package_fingerprints is None:
            package_versions = (
                dao.db.query(PackageVersion)
                .filter(PackageVersion.channel_name == channel_name)
                .filter(PackageVersion.platform == platform)
                .all()
            )

            package_fingerprints = {}
            for v in package_versions:
                info = json.loads(v.info)
                package_fingerprints[v.filename] = info.get(keyname)

        if keyname not in metadata:
            return None

        new_checksum = metadata[keyname]
        if package_name in package_fingerprints:
            existing_checksum = package_fingerprints[package_name]
            if existing_checksum is None:
                # missing checksum
                is_uptodate = None
            else:
                # compare  checksum
                is_uptodate = existing_checksum == new_checksum
        else:
            # missing package version
            is_uptodate = False

        return is_uptodate

    yield _func


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=4, max=10),
    after=after_log(logger, logging.WARNING),
)
def download_file(remote_repository, path_metadata):
    path, package_name, metadata = path_metadata
    try:
        f = remote_repository.open(path)
    except RemoteServerError as e:
        logger.error(f"remote server error when getting a file {path}")
        raise e
    except IncompleteRead as e:
        logger.error(f"Incomplete read for {path}")
        raise e

    logger.debug(f"Fetched file {path}")
    return f, package_name, metadata


def handle_repodata_package(
    channel,
    files_metadata,
    dao,
    auth,
    force,
    pkgstore,
    config,
):
    from quetz.main import pm

    channel_name = channel.name
    proxylist = channel.load_channel_metadata().get('proxylist', [])
    user_id = auth.assert_user()

    total_size = 0
    for file, package_name, metadata in files_metadata:
        parts = file.filename.rsplit("-", 2)
        if len(parts) != 3:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=f"package file name has wrong format {file.filename}",
            )
        else:
            package_name = parts[0]
        auth.assert_upload_file(channel_name, package_name)
        if force:
            auth.assert_overwrite_package_version(channel_name, package_name)

        # workaround for https://github.com/python/cpython/pull/3249
        if type(file.file) is SpooledTemporaryFile and not hasattr(file, "seekable"):
            file.file.seekable = file.file._file.seekable

        file.file.seek(0, os.SEEK_END)
        size = file.file.tell()
        total_size += size
        file.file.seek(0)

    dao.assert_size_limits(channel_name, total_size)

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=4, max=10),
        after=after_log(logger, logging.WARNING),
    )
    def _upload_package(file, channel_name, subdir):
        dest = os.path.join(subdir, file.filename)

        try:
            file.file.seek(0)
            logger.debug(
                f"uploading file {dest} from channel {channel_name} to package store"
            )
            pkgstore.add_package(file.file, channel_name, dest)

        except AttributeError as e:
            logger.error(f"Could not upload {file}, {file.filename}. {str(e)}")
            raise TryAgain

    pkgstore.create_channel(channel_name)
    nthreads = config.general_package_unpack_threads

    with TicToc("upload file without extracting"):
        with ThreadPoolExecutor(max_workers=nthreads) as executor:
            for file, package_name, metadata in files_metadata:
                if proxylist and package_name in proxylist:
                    # skip packages that should only ever be proxied
                    continue
                subdir = get_subdir_compat(metadata)
                executor.submit(_upload_package, file, channel_name, subdir)

    with TicToc("add versions to the db"):
        for file, package_name, metadata in files_metadata:
            version = create_version_from_metadata(
                channel_name, user_id, package_name, metadata, dao
            )
            condainfo = CondaInfo(file.file, package_name, lazy=True)
            pm.hook.post_add_package_version(version=version, condainfo=condainfo)
            file.file.close()


def initial_sync_mirror(
    channel_name: str,
    remote_repository: RemoteRepository,
    arch: str,
    dao: Dao,
    pkgstore: PackageStore,
    auth: authorization.Rules,
    includelist: List[str] = None,
    excludelist: List[str] = None,
    skip_errors: bool = True,
    use_repodata: bool = False,
):
    force = True  # needed for updating packages
    logger.info(
        f"Running channel mirroring {channel_name}/{arch} from {remote_repository.host}"
    )

    for repodata_fn in ["repodata_from_packages.json", "repodata.json"]:
        try:
            repo_file = remote_repository.open(os.path.join(arch, repodata_fn))
            repodata = json.load(repo_file.file)
            break
        except RemoteServerError:
            logger.error(
                f"can not get {repodata_fn} for channel {arch}/{channel_name}."
            )
            if repodata_fn == "repodata.json":
                logger.error(f"Giving up for {channel_name}/{arch}.")
                return
            else:
                logger.error("Trying next filename.")
                continue
        except json.JSONDecodeError:
            logger.error(
                f"repodata.json badly formatted for arch {arch}"
                f"in channel {channel_name}"
            )
            if repodata_fn == "repodata.json":
                return

    channel = dao.get_channel(channel_name)

    if not channel:
        logger.error(f"channel {channel_name} not found")
        return

    from quetz.main import handle_package_files

    packages = repodata.get("packages", {})

    version_methods = [
        _check_checksum(dao, channel_name, arch, "sha256"),
        _check_checksum(dao, channel_name, arch, "md5"),
    ]

    config = Config()
    max_batch_length = config.mirroring_batch_length
    max_batch_size = config.mirroring_batch_size
    # version_methods are context managers (for example, to update the db
    # after all packages have been checked), so we need to enter the context
    # for each
    any_updated = False
    with contextlib.ExitStack() as version_stack:
        version_checks = [
            version_stack.enter_context(method) for method in version_methods
        ]

        update_batch = []
        update_size = 0

        def handle_batch(update_batch):
            # i_batch += 1
            logger.info(f"Handling batch: {[p[1] for p in update_batch]}")
            if not update_batch:
                return False

            remote_packages = []
            remote_packages_with_metadata = []

            with ThreadPoolExecutor(
                max_workers=config.mirroring_num_parallel_downloads
            ) as executor:
                for f in executor.map(
                    download_file,
                    (remote_repository,) * len(update_batch),
                    update_batch,
                ):
                    if f is not None:
                        remote_packages.append(f[0])
                        remote_packages_with_metadata.append(f)

            try:
                if use_repodata:
                    handle_repodata_package(
                        channel,
                        remote_packages_with_metadata,
                        dao,
                        auth,
                        force,
                        pkgstore,
                        config,
                    )

                else:
                    handle_package_files(
                        channel,
                        remote_packages,
                        dao,
                        auth,
                        force,
                        is_mirror_op=True,
                    )
                return True

            except Exception as exc:
                logger.error(
                    f"could not process package {update_batch} from channel"
                    f"{channel_name} due to error {exc} of "
                    f"type {exc.__class__.__name__}"
                )
                if not skip_errors:
                    raise exc

            return False

        for package_name, metadata in packages.items():
            if check_package_membership(package_name, includelist, excludelist):
                path = os.path.join(arch, package_name)

                # try to find out whether it's a new package version

                is_uptodate = None
                for _check in version_checks:
                    is_uptodate = _check(package_name, metadata)
                    if is_uptodate is not None:
                        break

                # if package is up-to-date skip uploading file
                if is_uptodate:
                    continue
                else:
                    logger.debug(f"updating package {package_name} from {arch}")

                update_batch.append((path, package_name, metadata))
                update_size += metadata.get('size', 100_000)

            if len(update_batch) >= max_batch_length or update_size >= max_batch_size:
                logger.debug(f"Executing batch with {update_size}")
                any_updated |= handle_batch(update_batch)
                update_batch.clear()
                update_size = 0

        # handle final batch
        any_updated |= handle_batch(update_batch)

    if any_updated:
        indexing.update_indexes(dao, pkgstore, channel_name, subdirs=[arch])


def create_packages_from_channeldata(
    channel_name: str, user_id: bytes, channeldata: dict, dao: Dao
):
    packages = channeldata.get("packages", {})

    for package_name, metadata in packages.items():
        description = metadata.get("description", "")
        summary = metadata.get("summary", "")
        package_data = rest_models.Package(
            name=package_name,
            summary=summary,
            description=description,
        )

        try:
            package = dao.create_package(
                channel_name, package_data, user_id, role=authorization.OWNER
            )
        except DBError:
            # package already exists so skip it so we retrieve and update it
            package = dao.get_package(channel_name, package_name)
            package.description = description
            package.summary = summary
        package.url = metadata.get("home", "")
        package.platforms = ":".join(metadata.get("subdirs", []))
        package.channeldata = json.dumps(metadata)
        dao.db.commit()


def create_version_from_metadata(
    channel_name: str,
    user_id: bytes,
    package_file_name: str,
    package_data: dict,
    dao: Dao,
):
    package_name = package_data["name"]
    package = dao.get_package(channel_name, package_name)
    if not package:
        package_info = rest_models.Package(
            name=package_name,
            summary=package_data.get("summary", ""),
            description=package_data.get("description", ""),
        )
        dao.create_package(channel_name, package_info, user_id, "owner")

    pkg_format = "tarbz2" if package_file_name.endswith(".tar.bz2") else ".conda"
    version = dao.create_version(
        channel_name,
        package_name,
        pkg_format,
        get_subdir_compat(package_data),
        package_data["version"],
        int(package_data["build_number"]),
        package_data["build"],
        package_file_name,
        json.dumps(package_data),
        user_id,
        package_data["size"],
    )

    return version


def create_versions_from_repodata(
    channel_name: str, user_id: bytes, repodata: dict, dao: Dao
):
    packages = repodata.get("packages", {})
    for filename, metadata in packages.items():
        create_version_from_metadata(channel_name, user_id, filename, metadata, dao)


def synchronize_packages(
    channel_name: str,
    dao: Dao,
    pkgstore: PackageStore,
    auth: authorization.Rules,
    session: requests.Session,
    includelist: List[str] = None,
    excludelist: List[str] = None,
    use_repodata: bool = False,
):
    logger.info(f"executing synchronize_packages task in a process {os.getpid()}")

    new_channel = dao.get_channel(channel_name)

    if not new_channel:
        logger.error(f"channel {channel_name} not found")
        return

    host = new_channel.mirror_channel_url

    remote_repo = RemoteRepository(new_channel.mirror_channel_url, session)

    user_id = auth.assert_user()

    try:
        channel_data = remote_repo.open("channeldata.json").json()
        if use_repodata:
            create_packages_from_channeldata(channel_name, user_id, channel_data, dao)
        subdirs = channel_data.get("subdirs", [])
    except (RemoteFileNotFound, json.JSONDecodeError):
        subdirs = None
    except RemoteServerError:
        raise HTTPException(
            status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
            detail=f"Remote channel {host} unavailable",
        )
    # if no channel data use known architectures
    if subdirs is None:
        subdirs = KNOWN_SUBDIRS

    for arch in subdirs:
        initial_sync_mirror(
            new_channel.name,
            remote_repo,
            arch,
            dao,
            pkgstore,
            auth,
            includelist,
            excludelist,
            use_repodata=use_repodata,
        )



"""Added create_at and expire_at date to API key

Revision ID: ebe550f9fbbe
Revises: 0653794b6252
Create Date: 2021-01-22 22:38:05.693595

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'ebe550f9fbbe'
down_revision = '0653794b6252'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('api_keys', schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                'time_created',
                sa.Date(),
                nullable=False,
                server_default='2020-01-01',
            )
        )
        batch_op.add_column(sa.Column('expire_at', sa.Date(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table('api_keys', schema=None) as batch_op:
        batch_op.drop_column('expire_at')
        batch_op.drop_column('time_created')

    # ### end Alembic commands ###



"""add channel metadata

Revision ID: 3c3288034362
Revises: ea6eba9a9ffc
Create Date: 2021-02-09 18:45:13.198051

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '3c3288034362'
down_revision = 'ea6eba9a9ffc'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'channels',
        sa.Column('channel_metadata', sa.String(), server_default='{}', nullable=False),
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.drop_column('channels', 'channel_metadata')

    # ### end Alembic commands ###



import datetime
from unittest.mock import ANY

import pytest

from quetz.metrics.db_models import PackageVersionMetric


def test_get_package_list(package_version, package_name, channel_name, client):
    response = client.get("/api/dummylogin/bartosz")
    response = client.get(
        f"/api/channels/{channel_name}/packages/{package_name}/versions"
    )

    assert response.status_code == 200
    assert response.json() == [
        {
            "id": ANY,
            "channel_name": "my-channel",
            "package_name": "my-package",
            "platform": "linux-64",
            "version": "0.1",
            "build_string": "",
            "build_number": 0,
            "filename": "test-package-0.1-0.tar.bz2",
            "info": {},
            "uploader": {"name": "Bartosz", "avatar_url": "http:///avatar"},
            "time_created": ANY,
            "download_count": 0,
        }
    ]


def test_package_version_list_by_date(
    package_version, package_name, channel_name, client
):
    now = datetime.datetime.utcnow()
    later = now + datetime.timedelta(minutes=1)
    earlier = now - datetime.timedelta(minutes=1)

    response = client.get("/api/dummylogin/bartosz")
    response = client.get(
        f"/api/channels/{channel_name}/packages/{package_name}/versions"
        "?time_created__ge=" + later.isoformat()
    )

    assert response.status_code == 200
    assert response.json() == []

    response = client.get(
        f"/api/channels/{channel_name}/packages/{package_name}/versions"
        "?time_created__ge=" + earlier.isoformat()
    )
    assert response.status_code == 200
    assert len(response.json()) == 1


@pytest.mark.parametrize(
    "channel_role,expected_code",
    [("owner", 201), ("maintainer", 201), ("member", 403), (None, 403)],
)
def test_add_package_permissions(auth_client, public_channel, expected_code):
    response = auth_client.post(
        f"/api/channels/{public_channel.name}/packages",
        json={"name": "test-package", "summary": "none", "description": "none"},
    )

    assert response.status_code == expected_code


@pytest.mark.parametrize(
    "channel_role,expected_code",
    [("owner", 200), ("maintainer", 200), ("member", 403), (None, 403)],
)
def test_get_channel_members(auth_client, public_channel, expected_code):
    response = auth_client.get(f"/api/channels/{public_channel.name}/members")

    assert response.status_code == expected_code


@pytest.mark.parametrize(
    "role,expected_code",
    [
        ("owner", 201),
        ("maintainer", 201),
        ("member", 201),
        ("invalid", 422),
    ],
)
def test_post_channel_member(
    auth_client, public_channel, other_user, role, expected_code
):
    response = auth_client.post(
        f"/api/channels/{public_channel.name}/members",
        json={"username": other_user.username, "role": role},
    )

    assert response.status_code == expected_code

    if expected_code == 201:
        response = auth_client.get(f"/api/channels/{public_channel.name}/members")
        response.raise_for_status()
        for element in response.json():
            if element["user"]["username"] == other_user.username:
                assert element["role"] == role
                break
        else:
            raise RuntimeError(f"User '{other_user.username}' not found.")


def test_post_channel_member_unknown_user(auth_client, public_channel):
    response = auth_client.post(
        f"/api/channels/{public_channel.name}/members",
        json={"username": "unknown-user", "role": "member"},
    )

    assert response.status_code == 404


def test_delete_channel_member(auth_client, public_channel, other_user):
    auth_client.post(
        f"/api/channels/{public_channel.name}/members",
        json={"username": other_user.username, "role": "member"},
    )

    response = auth_client.delete(
        f"/api/channels/{public_channel.name}/members",
        params={"username": other_user.username},
    )

    assert response.status_code == 200


def test_delete_channel_member_no_member(auth_client, public_channel, other_user):
    response = auth_client.delete(
        f"/api/channels/{public_channel.name}/members",
        params={"username": other_user.username},
    )

    assert response.status_code == 404


def test_upload_wrong_file_type(auth_client, public_channel):
    files = {"files": ("my_package-0.1-0.tar.bz", b"dfdf")}
    response = auth_client.post(
        f"/api/channels/{public_channel.name}/files/", files=files
    )
    assert response.status_code == 400
    assert "not a bzip2 file" in response.json()['detail']


def test_increment_download_count(
    auth_client, public_channel, package_version, db, mocker
):
    def get_db(config):
        yield db

    mocker.patch("quetz.main.get_db", get_db)

    assert not db.query(PackageVersionMetric).one_or_none()

    with auth_client:
        response = auth_client.get(
            f"/get/{public_channel.name}/linux-64/test-package-0.1-0.tar.bz2"
        )
        assert response.status_code == 200

    metrics = (
        db.query(PackageVersionMetric)
        .filter(PackageVersionMetric.channel_name == public_channel.name)
        .filter(PackageVersionMetric.platform == package_version.platform)
        .filter(PackageVersionMetric.filename == package_version.filename)
        .all()
    )

    assert len(metrics) > 1
    assert metrics[0].count == 1
    db.refresh(package_version)
    assert package_version.download_count == 1

    with auth_client:
        response = auth_client.get(
            f"/get/{public_channel.name}/linux-64/test-package-0.1-0.tar.bz2"
        )

    assert response.status_code == 200

    db.refresh(metrics[0])
    assert metrics[0].count == 2
    db.refresh(package_version)
    assert package_version.download_count == 2



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.

from __future__ import annotations

import logging
from typing import List

from fastapi import APIRouter, Depends, HTTPException, Query
from fastapi import status as http_status

from quetz import authorization
from quetz.config import PAGINATION_LIMIT
from quetz.dao import Dao
from quetz.deps import get_dao, get_db, get_rules
from quetz.jobs import models as job_db_models
from quetz.rest_models import PaginatedResponse

from .models import JobStatus, TaskStatus
from .rest_models import Job, JobBase, JobUpdateModel, Task

api_router = APIRouter()

logger = logging.getLogger("quetz")


@api_router.get("/api/jobs", tags=["Jobs"], response_model=PaginatedResponse[Job])
def get_jobs(
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
    status: List[JobStatus] = Query([JobStatus.pending, JobStatus.running]),
    skip: int = 0,
    limit: int = PAGINATION_LIMIT,
):
    # if this is merged https://github.com/tiangolo/fastapi/issues/2077
    # we will be able to use non-exploded list, i.e., ?state=running,pending
    user_id = auth.assert_user()

    if auth.is_user_elevated(user_id):
        return dao.get_jobs(states=status, skip=skip, limit=limit)

    return dao.get_jobs(states=status, skip=skip, limit=limit, owner_id=user_id)


@api_router.post("/api/jobs", tags=["Jobs"], status_code=201, response_model=Job)
def create_job(
    job: JobBase,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    """create a new job"""
    user = auth.assert_user()
    # only admins can create jobs through /jobs API
    auth.assert_jobs(None)
    new_job = dao.create_job(user, job)
    return new_job


def get_job_or_fail(
    job_id: int,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
) -> job_db_models.Job:
    job = dao.get_job(job_id)

    if not job:
        raise HTTPException(
            status_code=http_status.HTTP_404_NOT_FOUND,
            detail=f"Job with id {job_id} not found",
        )

    auth.assert_jobs(job.owner_id)

    return job


@api_router.get("/api/jobs/{job_id}", tags=["Jobs"], response_model=Job)
def get_job(
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
    job: job_db_models.Job = Depends(get_job_or_fail),
):
    auth.assert_jobs(owner_id=job.owner_id)
    return job


@api_router.patch("/api/jobs/{job_id}", tags=["Jobs"])
def update_job(
    job_data: JobUpdateModel,
    db=Depends(get_db),
    job: job_db_models.Job = Depends(get_job_or_fail),
    auth: authorization.Rules = Depends(get_rules),
):
    """refresh job (re-run on new packages)"""
    auth.assert_jobs(owner_id=job.owner_id)
    job.status = job_data.status  # type: ignore

    if job_data.force and job.status in [
        JobStatus.running,
        JobStatus.pending,
    ]:
        # restart tasks that have already been run
        for task in job.tasks:
            task.status = "skipped"

    db.commit()


@api_router.get(
    "/api/jobs/{job_id}/tasks", tags=["Jobs"], response_model=PaginatedResponse[Task]
)
def get_tasks(
    job_id: int,
    dao: Dao = Depends(get_dao),
    status: List[TaskStatus] = Query(
        [TaskStatus.created, TaskStatus.pending, TaskStatus.running]
    ),
    auth: authorization.Rules = Depends(get_rules),
    skip: int = 0,
    limit: int = PAGINATION_LIMIT,
    job: job_db_models.Job = Depends(get_job_or_fail),
):
    auth.assert_jobs(owner_id=job.owner_id)
    return dao.get_tasks(job.id, status, skip, limit)


def get_router():
    return api_router



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.

import logging
import pickle
import re
import time
from datetime import datetime, timedelta
from typing import Dict, List

import sqlalchemy as sa
from sqlalchemy import Boolean
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import func
from sqlalchemy.sql.expression import FunctionElement

from quetz.db_models import PackageVersion
from quetz.jobs.models import ItemsSelection, Job, JobStatus, Task, TaskStatus
from quetz.jobs.rest_models import parse_job_manifest

logger = logging.getLogger('quetz.tasks')


class any_true(FunctionElement):
    inherit_cache = True
    name = "anytrue"
    type = Boolean()


@compiles(any_true, 'sqlite')
def sqlite_any(element, compiler, **kw):
    return f'max({compiler.process(element.clauses, **kw)})'


@compiles(any_true, 'postgresql')
def pg_any(element, compiler, **kw):
    return f'bool_or({compiler.process(element.clauses, **kw)})'


class all_true(FunctionElement):
    inherit_cache = True
    name = "alltrue"
    type = Boolean()


@compiles(all_true, 'sqlite')
def sqlite_all(element, compiler, **kw):
    return f'min({compiler.process(element.clauses, **kw)})'


@compiles(all_true, 'postgresql')
def pg_all(element, compiler, **kw):
    return f'bool_and({compiler.process(element.clauses, **kw)})'


def build_queue(job):
    job.status = JobStatus.queued


def parse_conda_spec(conda_spec: str):
    pattern = r'([a-zA-Z\*][^ =<>!~]*)([><!=~,\.0-9]+[0-9])?'
    exprs_list = re.findall(pattern, conda_spec)

    package_specs = []
    for name, versions in exprs_list:
        version_spec = None
        for spec_str in versions.split(','):
            if spec_str.startswith("=="):
                condition = ("eq", spec_str[2:])
            elif spec_str.startswith(">="):
                condition = ("gte", spec_str[2:])
            elif spec_str.startswith("<="):
                condition = ("lte", spec_str[2:])
            elif spec_str.startswith(">"):
                condition = ("gt", spec_str[1:])
            elif spec_str.startswith("<"):
                condition = ("lt", spec_str[1:])
            elif not spec_str:
                continue
            else:
                raise NotImplementedError("version operator not implemented")
            if version_spec:
                version_spec = ("and", version_spec, condition)
            else:
                version_spec = condition
        if "*" in name:
            dict_spec = {"package_name": ("like", name)}
        else:
            dict_spec = {"package_name": ("eq", name)}
        if version_spec:
            dict_spec["version"] = version_spec
        package_specs.append(dict_spec)
    return package_specs


def mk_sql_expr(dict_spec: List[Dict]):
    def _make_op(column, expr):
        op = expr[0]
        v = expr[1:]
        if op == 'eq':
            return column == v[0]
        elif op == 'in':
            return column.in_(v[0])
        elif op == 'lt':
            return column < v[0]
        elif op == 'gt':
            return column > v[0]
        elif op == 'gte':
            return column >= v[0]
        elif op == 'lte':
            return column <= v[0]
        elif op == "like":
            return column.ilike(v[0].replace("*", "%"))
        elif op == "and":
            left = _make_op(column, v[0])
            right = _make_op(column, v[1])
            return sa.and_(left, right)
        elif op == "or":
            left = _make_op(column, v[0])
            right = _make_op(column, v[1])
            return sa.or_(left, right)
        else:
            raise NotImplementedError(f"operator '{op}' not known")

    if not dict_spec:
        return False

    or_elements = []
    for el in dict_spec:
        and_elements = []
        for k, expr in el.items():
            column = getattr(PackageVersion, k)
            and_elements.append(_make_op(column, expr))

        expr = sa.and_(*and_elements)
        or_elements.append(expr)
    sql_expr = sa.or_(*or_elements)
    return sql_expr


def build_sql_from_package_spec(selector: str):
    dict_spec = parse_conda_spec(selector)
    sql_expr = mk_sql_expr(dict_spec)
    return sql_expr


class Supervisor:
    """Watches for new jobs and dispatches tasks."""

    def __init__(self, db, manager):
        self.db = db
        self.manager = manager
        self._process_cache = {}
        self._reset_tasks_after_restart()
        pass

    def _select_package_versions(self, job, force=False):
        db = self.db

        if job.items == ItemsSelection.all:
            if force:
                q = db.query(PackageVersion)
            else:
                existing_task = (
                    db.query(Task.package_version_id, Task.id.label("task_id"))
                    .filter(Task.job_id == job.id)
                    .filter(Task.status != TaskStatus.skipped)
                    .subquery()
                )
                q = (
                    db.query(PackageVersion)
                    .outerjoin(
                        existing_task,
                        PackageVersion.id == existing_task.c.package_version_id,
                    )
                    .filter(existing_task.c.task_id.is_(None))
                )
        else:
            raise NotImplementedError(f"selection {job.items} is not implemented")

        filter_expr = build_sql_from_package_spec(job.items_spec)
        q = q.filter(filter_expr)

        return q

    def run_jobs(self, job_id=None, force=False):
        now = datetime.utcnow()
        db = self.db
        jobs = db.query(Job).filter(Job.status == JobStatus.pending)
        if jobs.count():
            logger.info(f"Got pending jobs: {jobs.count()}")
        if job_id:
            jobs = jobs.filter(Job.id == job_id)
        for job in jobs:
            if job.start_at and job.start_at > now:
                continue

            should_repeat = (
                job.repeat_every_seconds
                and (job.updated + timedelta(seconds=job.repeat_every_seconds)) < now
            )

            if job.items_spec is not None:
                # it's a "package-version job"

                try:
                    force = force or should_repeat
                    q = self._select_package_versions(job, force=force)
                except Exception as e:
                    job.status = JobStatus.failed
                    logger.error(f"got error when parsing package spec: {e}")
                    continue

                task = None
                for version in q:
                    task = Task(job=job, package_version=version)
                    db.add(task)

                if not task and not job.repeat_every_seconds:
                    logger.info(
                        f"No new versions matching the package spec {job.items_spec}. "
                        f"Skipping job {job.id}."
                    )
                    job.status = JobStatus.success
                else:
                    job.status = JobStatus.running
                    job.updated = now

            else:
                # it's a "channel action job"
                if not job.tasks or should_repeat:
                    task = Task(job=job)
                    db.add(task)
                    job.updated = now
                    job.status = JobStatus.running

            db.commit()

    def add_task_to_queue(self, db, task, *args, **kwargs):
        """add task to the queue"""

        db = self.db
        manager = self.manager
        _process_cache = self._process_cache

        try:
            action_name = task.job.manifest.decode('ascii')
            action_func = parse_job_manifest(action_name)
        except UnicodeDecodeError:
            try:
                action_func = pickle.loads(task.job.manifest)
            except pickle.UnpicklingError:
                logger.error(
                    f"job {task.job_id} manifest contains non-ascii characters"
                )
                raise
        except ValueError:
            logger.error(f"job action {action_name} not known")
            raise

        task.status = TaskStatus.pending
        task.job.status = JobStatus.running
        db.add(task)
        db.commit()
        job = manager.execute(action_func, *args, task_id=task.id, **kwargs)
        _process_cache[task.id] = job
        return job

    def run_tasks(self):
        """dispatch tasks"""

        db = self.db

        tasks = db.query(Task).filter(Task.status == TaskStatus.created)
        task: Task
        if tasks.count():
            logger.info(f"Got pending tasks: {tasks.count()}")
        jobs = []
        for task in tasks:
            if not task.package_version:
                kwargs = {}
            else:
                kwargs = {
                    'package_version': {
                        "filename": task.package_version.filename,
                        "channel_name": task.package_version.channel_name,
                        "package_format": task.package_version.package_format,
                        "platform": task.package_version.platform,
                        "version": task.package_version.version,
                        "build_string": task.package_version.build_string,
                        "build_number": task.package_version.build_number,
                        "size": task.package_version.size,
                        "package_name": task.package_version.package_name,
                        "info": task.package_version.info,
                        "uploader_id": task.package_version.uploader_id,
                    }
                }
            try:
                job = self.add_task_to_queue(db, task, **kwargs)
                jobs.append(job)
            except Exception:
                logger.exception(f"task {task.id} failed due to error")
                task.status = TaskStatus.failed

        db.commit()
        return jobs

    def _reset_tasks_after_restart(self):
        # tasks lost after restart
        n_updated = (
            self.db.query(Task)
            .filter(Task.status.in_([TaskStatus.running, TaskStatus.pending]))
            .update({Task.status: TaskStatus.failed}, synchronize_session=False)
        )
        self.db.commit()

        if n_updated > 0:
            logger.warning(f"{n_updated} tasks set to failed due to supervisor restart")

    def _update_running_jobs(self):
        """Update status of running/pending jobs."""

        task_done = Task.status.in_([TaskStatus.failed, TaskStatus.success])
        running_job = Job.status.in_([JobStatus.running, JobStatus.pending])
        # we are using func.min/max to implement all/any aggregate functions
        results = (
            self.db.query(
                Task.job_id,
                func.min(Job.repeat_every_seconds),
                # flag if any task failed
                any_true(Task.status == TaskStatus.failed).label("failed"),
            )
            .outerjoin(Job)
            .filter(running_job)
            .group_by(Task.job_id)
            # select jobs where all tasks are finsihed
            .having(all_true(task_done))
            .all()
        )
        for job_id, repeat, failed in results:
            if repeat:
                # job with repeat non-null repeat column should be
                # kept runing until cancelled
                status = JobStatus.pending
            elif failed:
                status = JobStatus.failed
            else:
                status = JobStatus.success
            self.db.query(Job).filter(Job.id == job_id).update({Job.status: status})
        self.db.commit()

    def check_status(self):
        self._update_running_jobs()

    def run_once(self):
        self.run_jobs()
        self.run_tasks()
        self.check_status()

    def run(self):
        """main loop"""

        while True:
            self.run_once()
            time.sleep(5)



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.

# This code was largely taken from upstream conda
# File: conda/models/version.py
# https://github.com/conda/conda/blob/master/conda/models/version.py

# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause

import re
from itertools import zip_longest


class InvalidVersionSpec(ValueError):
    def __init__(self, invalid_spec: str, details: str):
        message = f"Invalid version '{invalid_spec}': {details}"
        super(InvalidVersionSpec, self).__init__(message)


version_check_re = re.compile(r'^[\*\.\+!_0-9a-z]+$')
version_split_re = re.compile(r'([0-9]+|[*]+|[^0-9*]+)')


class VersionOrder:
    """
    This class implements an order relation between version strings.
    Version strings can contain the usual alphanumeric characters
    (A-Za-z0-9), separated into components by dots and underscores. Empty
    segments (i.e. two consecutive dots, a leading/trailing underscore)
    are not permitted. An optional epoch number - an integer
    followed by '!' - can preceed the actual version string
    (this is useful to indicate a change in the versioning
    scheme itself). Version comparison is case-insensitive.

    Conda supports six types of version strings:

    * Release versions contain only integers, e.g. '1.0', '2.3.5'.
    * Pre-release versions use additional letters such as 'a' or 'rc',
      for example '1.0a1', '1.2.beta3', '2.3.5rc3'.
    * Development versions are indicated by the string 'dev',
      for example '1.0dev42', '2.3.5.dev12'.
    * Post-release versions are indicated by the string 'post',
      for example '1.0post1', '2.3.5.post2'.
    * Tagged versions have a suffix that specifies a particular
      property of interest, e.g. '1.1.parallel'. Tags can be added
      to any of the preceding four types. As far as sorting is concerned,
      tags are treated like strings in pre-release versions.
    * An optional local version string separated by '+' can be appended
      to the main (upstream) version string. It is only considered
      in comparisons when the main versions are equal, but otherwise
      handled in exactly the same manner.

    To obtain a predictable version ordering, it is crucial to keep the
    version number scheme of a given package consistent over time.
    Specifically,

    * version strings should always have the same number of components
      (except for an optional tag suffix or local version string),
    * letters/strings indicating non-release versions should always
      occur at the same position.

    Before comparison, version strings are parsed as follows:

    * They are first split into epoch, version number, and local version
      number at '!' and '+' respectively. If there is no '!', the epoch is
      set to 0. If there is no '+', the local version is empty.
    * The version part is then split into components at '.' and '_'.
    * Each component is split again into runs of numerals and non-numerals
    * Subcomponents containing only numerals are converted to integers.
    * Strings are converted to lower case, with special treatment for 'dev'
      and 'post'.
    * When a component starts with a letter, the fillvalue 0 is inserted
      to keep numbers and strings in phase, resulting in '1.1.a1' == 1.1.0a1'.
    * The same is repeated for the local version part.

    Examples:

        1.2g.beta15.rc  =>  [[0], [1], [2, 'g'], [0, 'beta', 15], [0, 'rc']]
        1!2.15.1_ALPHA  =>  [[1], [2], [15], [1, '_alpha']]

    The resulting lists are compared lexicographically, where the following
    rules are applied to each pair of corresponding subcomponents:

    * integers are compared numerically
    * strings are compared lexicographically, case-insensitive
    * strings are smaller than integers, except
    * 'dev' versions are smaller than all corresponding versions of other types
    * 'post' versions are greater than all corresponding versions of other types
    * if a subcomponent has no correspondent, the missing correspondent is
      treated as integer 0 to ensure '1.1' == '1.1.0'.

    The resulting order is:

           0.4
         < 0.4.0
         < 0.4.1.rc
        == 0.4.1.RC   # case-insensitive comparison
         < 0.4.1
         < 0.5a1
         < 0.5b3
         < 0.5C1      # case-insensitive comparison
         < 0.5
         < 0.9.6
         < 0.960923
         < 1.0
         < 1.1dev1    # special case 'dev'
         < 1.1_       # appended underscore is special case for openssl-like versions
         < 1.1a1
         < 1.1.0dev1  # special case 'dev'
        == 1.1.dev1   # 0 is inserted before string
         < 1.1.a1
         < 1.1.0rc1
         < 1.1.0
        == 1.1
         < 1.1.0post1 # special case 'post'
        == 1.1.post1  # 0 is inserted before string
         < 1.1post1   # special case 'post'
         < 1996.07.12
         < 1!0.4.1    # epoch increased
         < 1!3.1.1.6
         < 2!0.4.1    # epoch increased again

    Some packages (most notably openssl) have incompatible version conventions.
    In particular, openssl interprets letters as version counters rather than
    pre-release identifiers. For openssl, the relation

      1.0.1 < 1.0.1a  =>  False  # should be true for openssl

    holds, whereas conda packages use the opposite ordering. You can work-around
    this problem by appending an underscore to plain version numbers:

      1.0.1_ < 1.0.1a =>  True   # ensure correct ordering for openssl
    """

    def __init__(self, vstr: str):
        # version comparison is case-insensitive
        version = vstr.strip().rstrip().lower()
        # basic validity checks
        if version == '':
            raise InvalidVersionSpec(vstr, "empty version string")
        invalid = not version_check_re.match(version)
        if invalid and '-' in version and '_' not in version:
            # Allow for dashes as long as there are no underscores
            # as well, by converting the former to the latter.
            version = version.replace('-', '_')
            invalid = not version_check_re.match(version)
        if invalid:
            raise InvalidVersionSpec(vstr, "invalid character(s)")

        # when fillvalue ==  0  =>  1.1 == 1.1.0
        # when fillvalue == -1  =>  1.1  < 1.1.0
        self.norm_version = version
        self.fillvalue = 0

        # find epoch
        split_epoch = version.split('!')
        if len(split_epoch) == 1:
            # epoch not given => set it to '0'
            epoch = ['0']
        elif len(split_epoch) == 2:
            # epoch given, must be an integer
            if not split_epoch[0].isdigit():
                raise InvalidVersionSpec(vstr, "epoch must be an integer")
            epoch = [split_epoch[0]]
            version = split_epoch[1]
        else:
            raise InvalidVersionSpec(vstr, "duplicated epoch separator '!'")

        # find local version string
        split_local = version.split('+')
        if len(split_local) == 1:
            # no local version
            self.local = []
        elif len(split_local) == 2:
            # local version given
            self.local = split_local[1].replace('_', '.').split('.')
            version = split_local[0]
        else:
            raise InvalidVersionSpec(vstr, "duplicated local version separator '+'")

        # split version
        if version[-1] == "_":
            # If the last character of version is "-" or "_", don't split that out
            # individually. Implements the instructions for openssl-like versions
            # > You can work-around this problem by appending a dash to plain version
            #   numbers
            split_version = version[:-1].replace('_', '.').split('.')
            split_version[-1] += "_"
        else:
            split_version = version.replace('_', '.').split('.')
        self.version = epoch + split_version

        # split components into runs of numerals and non-numerals,
        # convert numerals to int, handle special strings
        for v in (self.version, self.local):
            for k in range(len(v)):
                c = version_split_re.findall(v[k])
                if not c:
                    raise InvalidVersionSpec(vstr, "empty version component")
                for j in range(len(c)):
                    if c[j].isdigit():
                        c[j] = int(c[j])
                    elif c[j] == 'post':
                        # ensure number < 'post' == infinity
                        c[j] = float('inf')
                    elif c[j] == 'dev':
                        # ensure '*' < 'DEV' < '_' < 'a' < number
                        # by upper-casing (all other strings are lower case)
                        c[j] = 'DEV'
                if v[k][0].isdigit():
                    v[k] = c  # type: ignore
                else:
                    # components shall start with a number to keep numbers and
                    # strings in phase => prepend fillvalue
                    v[k] = [self.fillvalue] + c  # type: ignore

    def __str__(self):
        return self.norm_version

    def __repr__(self):
        return f'{self.__class__.__name__}("{self}")'

    def _eq(self, t1, t2):
        for v1, v2 in zip_longest(t1, t2, fillvalue=[]):
            for c1, c2 in zip_longest(v1, v2, fillvalue=self.fillvalue):
                if c1 != c2:
                    return False
        return True

    def __eq__(self, other):
        return self._eq(self.version, other.version) and self._eq(
            self.local, other.local
        )

    def startswith(self, other):
        # Tests if the version lists match up to the last element in "other".
        if other.local:
            if not self._eq(self.version, other.version):
                return False
            t1 = self.local
            t2 = other.local
        else:
            t1 = self.version
            t2 = other.version
        nt = len(t2) - 1
        if not self._eq(t1[:nt], t2[:nt]):
            return False
        v1 = [] if len(t1) <= nt else t1[nt]
        v2 = t2[nt]
        nt = len(v2) - 1
        if not self._eq([v1[:nt]], [v2[:nt]]):
            return False
        c1 = self.fillvalue if len(v1) <= nt else v1[nt]
        c2 = v2[nt]
        if isinstance(c2, str):
            return isinstance(c1, str) and c1.startswith(c2)
        return c1 == c2

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        for t1, t2 in zip([self.version, self.local], [other.version, other.local]):
            for v1, v2 in zip_longest(t1, t2, fillvalue=[]):
                for c1, c2 in zip_longest(v1, v2, fillvalue=self.fillvalue):
                    if c1 == c2:
                        continue
                    elif isinstance(c1, str):
                        if not isinstance(c2, str):
                            # str < int
                            return True
                    elif isinstance(c2, str):
                        # not (int < str)
                        return False
                    # c1 and c2 have the same type
                    return c1 < c2
        # self == other
        return False

    def __gt__(self, other):
        return other < self

    def __le__(self, other):
        return not (other < self)

    def __ge__(self, other):
        return not (self < other)



import os

from alembic import context
from alembic.config import Config as AlembicConfig
from sqlalchemy import engine_from_config, pool

from quetz.config import Config
from quetz.db_models import Base

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config: AlembicConfig = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
# fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata

target_metadata = Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def get_url():
    db_path = config.get_main_option("sqlalchemy.url")
    if not db_path:
        config_path = context.get_x_argument(as_dictionary=True).get('quetzConfig')
        deployment_path = os.path.split(config_path)[0]
        quetz_config = Config(config_path)
        db_path = quetz_config.sqlalchemy_database_url
        abs_path = os.path.abspath(deployment_path)

        db_path = db_path.replace("sqlite:///.", f"sqlite:///{abs_path}")
    return db_path


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = get_url()

    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    configuration = config.get_section(config.config_ini_section)
    configuration["sqlalchemy.url"] = get_url()
    connection = config.attributes.get('connection', None)

    if connection is None:
        engine = engine_from_config(
            configuration,
            prefix="sqlalchemy.",
            poolclass=pool.NullPool,
        )
        connection = engine.connect()

    context.configure(
        connection=connection,
        target_metadata=target_metadata,
        compare_type=True,
        render_as_batch=True,
    )

    with context.begin_transaction():
        context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



class PackageError(Exception):
    def __init__(self, detail: str) -> None:
        self.detail = detail

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(detail={self.detail!r})"



import os

from prometheus_client import (
    CONTENT_TYPE_LATEST,
    REGISTRY,
    CollectorRegistry,
    generate_latest,
)
from prometheus_client.multiprocess import MultiProcessCollector
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

from .middleware import PrometheusMiddleware


def metrics(request: Request) -> Response:
    if "prometheus_multiproc_dir" in os.environ:
        registry = CollectorRegistry()
        MultiProcessCollector(registry)
    else:
        registry = REGISTRY

    return Response(generate_latest(registry), media_type=CONTENT_TYPE_LATEST)


def init(app: ASGIApp):
    app.add_middleware(PrometheusMiddleware)
    app.add_route("/metricsp", metrics)



"""scheduling spec for jobs

Revision ID: cddba8e6e639
Revises: 0a0ab48887ab
Create Date: 2021-03-02 11:26:04.330580

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'cddba8e6e639'
down_revision = '0a0ab48887ab'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.add_column(
            sa.Column('repeat_every_seconds', sa.Integer(), nullable=True)
        )
        batch_op.add_column(sa.Column('start_at', sa.DateTime(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.drop_column('start_at')
        batch_op.drop_column('repeat_every_seconds')

    # ### end Alembic commands ###



"""use BigInteger for size

Revision ID: 53f81aba78ce
Revises: 30241b33d849
Create Date: 2021-01-08 22:09:48.268270

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '53f81aba78ce'
down_revision = '30241b33d849'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    if op.get_context().dialect.name == 'postgresql':
        op.alter_column(
            'channels',
            'size',
            existing_type=sa.INTEGER(),
            type_=sa.BigInteger(),
            existing_nullable=True,
        )
        op.alter_column(
            'channels',
            'size_limit',
            existing_type=sa.INTEGER(),
            type_=sa.BigInteger(),
            existing_nullable=True,
        )

        op.alter_column(
            'package_versions',
            'size',
            existing_type=sa.INTEGER(),
            type_=sa.BigInteger(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    if op.get_context().dialect.name == 'postgresql':
        op.alter_column(
            'package_versions',
            'size',
            existing_type=sa.BigInteger(),
            type_=sa.INTEGER(),
            existing_nullable=True,
        )

        op.alter_column(
            'channels',
            'size_limit',
            existing_type=sa.BigInteger(),
            type_=sa.INTEGER(),
            existing_nullable=True,
        )
        op.alter_column(
            'channels',
            'size',
            existing_type=sa.BigInteger(),
            type_=sa.INTEGER(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###



"""add Email table for email addresses

Revision ID: d212023a8e0b
Revises: cddba8e6e639
Create Date: 2021-09-07 18:14:30.387156

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'd212023a8e0b'
down_revision = 'cddba8e6e639'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'emails',
        sa.Column('provider', sa.String(), nullable=False),
        sa.Column('identity_id', sa.String(), nullable=False),
        sa.Column('email', sa.String(), nullable=False),
        sa.Column('user_id', sa.LargeBinary(length=16), nullable=True),
        sa.Column('verified', sa.Boolean(), nullable=True),
        sa.Column('primary', sa.Boolean(), nullable=True),
        sa.ForeignKeyConstraint(
            ['provider', 'identity_id'],
            ['identities.provider', 'identities.identity_id'],
        ),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['users.id'],
        ),
        sa.PrimaryKeyConstraint('provider', 'identity_id', 'email'),
        sa.UniqueConstraint('email'),
    )
    with op.batch_alter_table('emails', schema=None) as batch_op:
        batch_op.create_index(
            'email_index', ['provider', 'identity_id', 'email'], unique=True
        )


def downgrade():
    with op.batch_alter_table('emails', schema=None) as batch_op:
        batch_op.drop_index('email_index')

    op.drop_table('emails')



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.
import asyncio
import datetime
import hashlib
import json
import logging
import os
import re
import time
import uuid
from collections import Counter
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from email.utils import formatdate
from tempfile import SpooledTemporaryFile, TemporaryFile
from typing import List, Optional, Tuple, Type

import pydantic
import requests
from fastapi import (
    APIRouter,
    BackgroundTasks,
    Depends,
    FastAPI,
    File,
    Form,
    Header,
    HTTPException,
    Request,
    UploadFile,
    responses,
    status,
)
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import RedirectResponse, StreamingResponse
from importlib_metadata import entry_points
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session
from starlette.concurrency import run_in_threadpool
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.httpsredirect import HTTPSRedirectMiddleware
from starlette.middleware.sessions import SessionMiddleware
from tenacity import (
    after_log,
    retry,
    retry_if_result,
    stop_after_attempt,
    wait_exponential,
)

from quetz import (
    authorization,
    db_models,
    errors,
    exceptions,
    frontend,
    metrics,
    rest_models,
)
from quetz.authentication import AuthenticatorRegistry, BaseAuthenticator
from quetz.authentication import github as auth_github
from quetz.authentication import gitlab as auth_gitlab
from quetz.authentication import google as auth_google
from quetz.authentication.azuread import AzureADAuthenticator
from quetz.authentication.jupyterhub import JupyterhubAuthenticator
from quetz.authentication.pam import PAMAuthenticator
from quetz.config import PAGINATION_LIMIT, Config, configure_logger, get_plugin_manager
from quetz.dao import Dao
from quetz.deps import (
    ChannelChecker,
    get_channel_allow_proxy,
    get_channel_or_fail,
    get_config,
    get_dao,
    get_db,
    get_package_or_fail,
    get_remote_session,
    get_rules,
    get_session,
    get_tasks_worker,
)
from quetz.jobs import api as jobs_api
from quetz.jobs import rest_models as jobs_rest
from quetz.metrics import api as metrics_api
from quetz.metrics.middleware import DOWNLOAD_COUNT, UPLOAD_COUNT
from quetz.rest_models import ChannelActionEnum, CPRole
from quetz.tasks import indexing
from quetz.tasks.common import Task
from quetz.tasks.mirror import RemoteRepository, download_remote_file
from quetz.utils import (
    TicToc,
    background_task_wrapper,
    generate_random_key,
    parse_query,
)

from .condainfo import CondaInfo

app = FastAPI()

config = Config()

configure_logger(config)

logger = logging.getLogger("quetz")

app.add_middleware(
    SessionMiddleware,
    secret_key=config.session_secret,
    https_only=config.session_https_only,
)

if config.general_redirect_http_to_https:
    logger.info("Configuring http to https redirect ")
    app.add_middleware(HTTPSRedirectMiddleware)

metrics.init(app)

if config.configured_section("cors"):
    logger.info("Configuring CORS with ")
    logger.info(f"allow_origins     = {config.cors_allow_origins}")
    logger.info(f"allow_credentials = {config.cors_allow_credentials}")
    logger.info(f"allow_methods     = {config.cors_allow_methods}")
    logger.info(f"allow_headers     = {config.cors_allow_headers}")
    app.add_middleware(
        CORSMiddleware,
        allow_origins=config.cors_allow_origins,
        allow_credentials=config.cors_allow_credentials,
        allow_methods=config.cors_allow_methods,
        allow_headers=config.cors_allow_headers,
    )

# global variables for batching download counts

download_counts: Counter = Counter()

DOWNLOAD_INCREMENT_DELAY_SECONDS = 10
DOWNLOAD_INCREMENT_MAX_DOWNLOADS = 50


class CondaTokenMiddleware(BaseHTTPMiddleware):
    """Removes /t/<QUETZ_API_KEY> prefix, adds QUETZ_APY_KEY to the headers and passes
    on the rest of the path to be routed."""

    def __init__(self, app):
        super().__init__(app)
        self.token_pattern = re.compile("^/t/([^/]+?)/")

    async def dispatch(self, request, call_next):
        path = request.scope["path"]
        match = self.token_pattern.search(path)
        if match:
            prefix_length = len(match.group(0)) - 1
            new_path = path[prefix_length:]
            api_key = match.group(1)
            request.scope["path"] = new_path
            request.scope["headers"].append((b"x-api-key", api_key.encode()))

        response = await call_next(request)

        return response


app.add_middleware(CondaTokenMiddleware)

pkgstore = config.get_package_store()

# authenticators


builtin_authenticators: List[Type[BaseAuthenticator]] = [
    authenticator
    for authenticator in [
        auth_github.GithubAuthenticator,
        auth_gitlab.GitlabAuthenticator,
        auth_google.GoogleAuthenticator,
        JupyterhubAuthenticator,
        PAMAuthenticator,
        AzureADAuthenticator,
    ]
    if authenticator is not None
]

plugin_authenticators: List[Type[BaseAuthenticator]] = [
    ep.load() for ep in entry_points().select(group='quetz.authenticator')
]


auth_registry = AuthenticatorRegistry()
auth_registry.set_router(app)

for auth_cls in builtin_authenticators + plugin_authenticators:
    auth_obj = auth_cls(config)
    if auth_obj.is_enabled:
        auth_registry.register(auth_obj)

# other routers

pm = get_plugin_manager()
api_router = APIRouter()
plugin_routers = pm.hook.register_router()

for router in plugin_routers:
    app.include_router(router)
app.include_router(jobs_api.get_router())
app.include_router(metrics_api.get_router())


# helper functions
async def check_token_revocation(session):
    valid = True
    identity_provider = session.get("identity_provider")

    if identity_provider is None:
        valid = False
    elif identity_provider != "dummy":
        auth_obj = auth_registry.enabled_authenticators[identity_provider]
        valid = await auth_obj.validate_token(session.get("token"))
    if not valid:
        logout(session)
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not logged in",
        )


def logout(session):
    session.pop("user_id", None)
    session.pop("identity_provider", None)
    session.pop("token", None)


# custom exception handlers
@app.exception_handler(errors.ValidationError)
async def unicorn_exception_handler(request: Request, exc: errors.ValidationError):
    return responses.JSONResponse(
        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, content={"detail": str(exc)}
    )


# endpoints
@app.route("/auth/logout")
async def route_logout(request):
    logout(request.session)
    return RedirectResponse("/")


@api_router.get("/dummylogin/{username}", tags=["dev"])
def dummy_login(
    username: str, dao: Dao = Depends(get_dao), session=Depends(get_session)
):
    user = dao.get_user_by_username(username)

    logout(session)
    session["user_id"] = str(uuid.UUID(bytes=user.id))

    session["identity_provider"] = "dummy"
    return RedirectResponse("/")


@api_router.get("/me", response_model=rest_models.Profile, tags=["users"])
async def me(
    session: dict = Depends(get_session),
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    """Returns your quetz profile"""

    # Check if token is still valid
    await check_token_revocation(session)

    user_id = auth.assert_user()

    profile = dao.get_profile(user_id)
    if profile:
        profile.user.id = str(uuid.UUID(bytes=profile.user.id))

    return profile


def get_users_handler(dao, q, auth, skip, limit):
    user_id = auth.assert_user()

    results = dao.get_users(skip, limit, q)

    user_list = results["result"] if "result" in results else results

    if not auth.is_user_elevated(user_id):
        append_user = None
        for user in user_list:
            if user.id == user_id:
                append_user = user
        user_list.clear()
        if append_user:
            user_list.append(append_user)

    return results


@api_router.get("/users", response_model=List[rest_models.User], tags=["users"])
def get_users(
    dao: Dao = Depends(get_dao),
    q: str = None,
    auth: authorization.Rules = Depends(get_rules),
):
    return get_users_handler(dao, q, auth, 0, -1)


@api_router.get(
    "/paginated/users",
    response_model=rest_models.PaginatedResponse[rest_models.User],
    tags=["users"],
)
def get_paginated_users(
    dao: Dao = Depends(get_dao),
    skip: int = 0,
    limit: int = PAGINATION_LIMIT,
    q: str = None,
    auth: authorization.Rules = Depends(get_rules),
):
    return get_users_handler(dao, q, auth, skip, limit)


@api_router.get("/users/{username}", response_model=rest_models.User, tags=["users"])
def get_user(
    username: str,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    user = dao.get_user_by_username(username)

    if not user or not user.profile:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail=f"User {username} not found"
        )

    auth.assert_read_user_data(user.id)

    return user


@api_router.get(
    "/users/{username}/channels",
    response_model=List[rest_models.ChannelRole],
    tags=["users"],
)
def get_user_channels(
    username: str,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    return list_user_channels(username, dao, auth, 0, -1)


@api_router.get(
    "/users/{username}/packages",
    response_model=List[rest_models.PackageRole],
    tags=["users"],
)
def get_user_packages(
    username: str,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    return list_user_packages(username, dao, auth, 0, -1)


@api_router.get(
    "/paginated/users/{username}/channels",
    response_model=rest_models.PaginatedResponse[rest_models.ChannelRole],
    tags=["users"],
)
def get_paginated_user_channels(
    username: str,
    skip: int = 0,
    limit: int = PAGINATION_LIMIT,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    return list_user_channels(username, dao, auth, skip, limit)


@api_router.get(
    "/paginated/users/{username}/packages",
    response_model=rest_models.PaginatedResponse[rest_models.PackageRole],
    tags=["users"],
)
def get_paginated_user_packages(
    username: str,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
    skip: int = 0,
    limit: int = PAGINATION_LIMIT,
):
    return list_user_packages(username, dao, auth, skip, limit)


def list_user_packages(
    username: str,
    dao: Dao,
    auth: authorization.Rules,
    skip: int,
    limit: int,
):
    user = dao.get_user_by_username(username)

    if not user or not user.profile:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail=f"User {username} not found"
        )

    auth.assert_read_user_data(user.id)

    return dao.get_user_packages(skip, limit, user.id)


def list_user_channels(
    username: str, dao: Dao, auth: authorization.Rules, skip: int, limit: int
):
    user = dao.get_user_by_username(username)

    if not user or not user.profile:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail=f"User {username} not found"
        )

    auth.assert_read_user_data(user.id)

    channels = dao.get_user_channels_with_role(skip, limit, user.id)

    return channels


@api_router.delete("/users/{username}", tags=["users"])
def delete_user(
    username: str,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    user = dao.get_user_by_username(username)

    auth.assert_delete_user(user.id)
    dao.delete_user(user.id)


@api_router.get(
    "/users/{username}/role",
    response_model=rest_models.UserRole,
    tags=["users"],
)
def get_user_role(
    username: str,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    user = dao.get_user_by_username(username)

    if not user:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail=f"User {username} not found"
        )

    auth.assert_read_user_data(user.id)

    return {"role": user.role}


@api_router.put("/users/{username}/role", tags=["users"])
def set_user_role(
    username: str,
    role: rest_models.UserRole,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    user = dao.get_user_by_username(username)

    if not user:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail=f"User {username} not found"
        )

    auth.assert_assign_user_role(role.role)

    dao.set_user_role(username, role=role.role)


@api_router.get(
    "/channels", response_model=List[rest_models.ChannelExtra], tags=["channels"]
)
def get_channels(
    public: bool = True,
    dao: Dao = Depends(get_dao),
    q: str = None,
    auth: authorization.Rules = Depends(get_rules),
):
    """List all channels"""

    user_id = auth.get_user()
    return dao.get_channels(0, -1, q, user_id, include_public=public)


@api_router.get(
    "/paginated/channels",
    response_model=rest_models.PaginatedResponse[rest_models.ChannelExtra],
    tags=["channels"],
)
def get_paginated_channels(
    dao: Dao = Depends(get_dao),
    skip: int = 0,
    limit: int = PAGINATION_LIMIT,
    public: bool = True,
    q: str = None,
    auth: authorization.Rules = Depends(get_rules),
):
    """List all channels, as a paginated response"""
    user_id = auth.get_user()
    return dao.get_channels(skip, limit, q, user_id, include_public=public)


@api_router.get(
    "/channels/{channel_name}",
    response_model=rest_models.ChannelExtra,
    tags=["channels"],
)
def get_channel(channel: db_models.Channel = Depends(get_channel_allow_proxy)):
    return channel


@api_router.post(
    "/channels/{channel_name}/mirrors",
    status_code=201,
    tags=["channels"],
)
def post_channel_mirror(
    request: Request,
    mirror: rest_models.ChannelMirrorBase,
    channel_name: str,
    channel: db_models.Channel = Depends(get_channel_or_fail),
    auth: authorization.Rules = Depends(get_rules),
    dao: Dao = Depends(get_dao),
    remote_session: requests.Session = Depends(get_remote_session),
):
    auth.assert_register_mirror(channel_name)

    logger.debug(f"registering mirror {mirror.url}")

    if not mirror.api_endpoint:
        mirror.api_endpoint = mirror.url.replace("get", "api/channels")

    if not mirror.metrics_endpoint:
        mirror.metrics_endpoint = mirror.url.replace("get", "metrics/channels")

    # check api response
    response = remote_session.get(mirror.api_endpoint)

    if response.status_code != 200:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"could not connect to remote repository {mirror.url}",
        )
    response_data = response.json()

    try:
        mirrored_server = response_data["mirror_channel_url"]
    except KeyError:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="mirror server is not quetz server",
        )

    if not mirrored_server:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"{mirror.url} is not a mirror server",
        )

    dao.create_channel_mirror(
        channel_name, mirror.url, mirror.api_endpoint, mirror.metrics_endpoint
    )

    logger.info(f"successfully registered mirror {mirror.url}")


@api_router.get(
    "/channels/{channel_name}/mirrors",
    response_model=List[rest_models.ChannelMirror],
    tags=["channels"],
)
def get_channel_mirrors(
    channel_name: str,
    channel: db_models.Channel = Depends(get_channel_or_fail),
    auth: authorization.Rules = Depends(get_rules),
    dao: Dao = Depends(get_dao),
):
    return channel.mirrors


@api_router.delete(
    "/channels/{channel_name}/mirrors/{mirror_id}",
    tags=["channels"],
)
def delete_channel_mirror(
    channel_name: str,
    mirror_id: str,
    channel: db_models.Channel = Depends(get_channel_or_fail),
    auth: authorization.Rules = Depends(get_rules),
    dao: Dao = Depends(get_dao),
):
    auth.assert_unregister_mirror(channel_name)
    dao.delete_channel_mirror(channel_name, mirror_id)


@api_router.delete(
    "/channels/{channel_name}",
    tags=["channels"],
)
def delete_channel(
    channel: db_models.Channel = Depends(get_channel_allow_proxy),
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    auth.assert_delete_channel(channel)
    dao.delete_channel(channel.name)
    try:
        pkgstore.remove_channel(channel.name)
    except FileNotFoundError:
        logger.warning(
            f"trying to remove non-existent package store for channel {channel.name}"
        )


@api_router.put(
    "/channels/{channel_name}/actions", tags=["channels"], response_model=jobs_rest.Job
)
def put_mirror_channel_actions(
    action: rest_models.ChannelAction,
    channel: db_models.Channel = Depends(get_channel_allow_proxy),
    dao: Dao = Depends(get_dao),
    task: Task = Depends(get_tasks_worker),
):
    new_job = task.execute_channel_action(
        action.action,
        channel,
        start_at=action.start_at,
        repeat_every_seconds=action.repeat_every_seconds,
    )
    return new_job


@api_router.post("/channels", status_code=201, tags=["channels"])
def post_channel(
    request: Request,
    new_channel: rest_models.Channel,
    background_tasks: BackgroundTasks,
    mirror_api_key: Optional[str] = None,
    register_mirror: bool = False,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
    task: Task = Depends(get_tasks_worker),
    config=Depends(get_config),
    session: requests.Session = Depends(get_remote_session),
):
    user_id = auth.assert_user()

    existing_channel = dao.get_channel(new_channel.name)

    if existing_channel:
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Channel {new_channel.name} exists",
        )

    if not new_channel.mirror_channel_url:
        auth.assert_create_channel()

    is_mirror = new_channel.mirror_channel_url and new_channel.mirror_mode == "mirror"

    is_proxy = new_channel.mirror_channel_url and new_channel.mirror_mode == "proxy"

    if is_mirror:
        auth.assert_create_mirror_channel()

    if is_proxy:
        auth.assert_create_proxy_channel()

    if new_channel.actions is None:
        if is_mirror:
            actions = [ChannelActionEnum.synchronize_repodata]
        else:
            actions = []
    else:
        actions = new_channel.actions

    includelist = new_channel.metadata.includelist
    excludelist = new_channel.metadata.excludelist

    if includelist is not None and excludelist is not None:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Cannot use both `includelist` and `excludelist` together.",
        )

    user_attrs = new_channel.dict(exclude_unset=True)

    if "size_limit" in user_attrs:
        auth.assert_set_channel_size_limit()
        size_limit = new_channel.size_limit
    else:
        if config.configured_section("quotas"):
            size_limit = config.quotas_channel_quota
        else:
            size_limit = None

    channel = dao.create_channel(new_channel, user_id, authorization.OWNER, size_limit)
    pkgstore.create_channel(new_channel.name)
    if not is_proxy:
        indexing.update_indexes(dao, pkgstore, new_channel.name)

    # register mirror
    if is_mirror and register_mirror:
        mirror_url = str(new_channel.mirror_channel_url)
        mirror_url = mirror_url.replace("get", "api/channels")
        headers = {"x-api-key": mirror_api_key} if mirror_api_key else {}
        api_endpoint = str(request.url.replace(query=None)) + '/' + new_channel.name
        request.url
        response = session.post(
            mirror_url + '/mirrors',
            json={
                "url": api_endpoint.replace("api/channels", "get"),
                "api_endpoint": api_endpoint,
                "metrics_endpoint": api_endpoint.replace("api", "metrics"),
            },
            headers=headers,
        )
        if response.status_code != 201:
            logger.warning(f"could not register mirror due to error {response.text}")

    for action in actions:
        task.execute_channel_action(
            action,
            channel,
        )


@api_router.patch(
    "/channels/{channel_name}",
    status_code=200,
    tags=["channels"],
    response_model=rest_models.ChannelBase,
)
def patch_channel(
    channel_data: rest_models.Channel,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
    channel: db_models.Channel = Depends(get_channel_or_fail),
    db=Depends(get_db),
):
    auth.assert_update_channel_info(channel.name)

    user_attrs = channel_data.dict(exclude_unset=True)

    if "size_limit" in user_attrs:
        auth.assert_set_channel_size_limit()

    changeable_attrs = ["private", "size_limit", "metadata", "ttl"]

    for attr_ in user_attrs.keys():
        if attr_ not in changeable_attrs:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail=f"attribute '{attr_}' of channel can not be changed",
            )

    for attr_, value_ in user_attrs.items():
        if attr_ == "metadata":
            metadata = channel.load_channel_metadata()
            metadata.update(value_)
            setattr(channel, "channel_metadata", json.dumps(metadata))
        else:
            setattr(channel, attr_, value_)
    db.commit()

    return channel


@api_router.get(
    "/channels/{channel_name}/packages",
    response_model=List[rest_models.Package],
    tags=["packages"],
)
def get_packages(
    channel: db_models.Channel = Depends(get_channel_or_fail),
    dao: Dao = Depends(get_dao),
    q: Optional[str] = None,
):
    """
    Retrieve all packages in a channel, optionally matching a query `q`.
    """
    res = dao.get_packages(channel.name, 0, -1, q)
    return res


@api_router.get(
    "/paginated/channels/{channel_name}/packages",
    response_model=rest_models.PaginatedResponse[rest_models.Package],
    tags=["packages"],
)
def get_paginated_packages(
    channel: db_models.Channel = Depends(get_channel_or_fail),
    dao: Dao = Depends(get_dao),
    skip: int = 0,
    limit: int = PAGINATION_LIMIT,
    q: Optional[str] = None,
    order_by: Optional[str] = None,
):
    """
    Retrieve all packages in a channel.
    A limit of -1 returns an unpaginated result with all packages. Otherwise, pagination
    is applied.
    """

    return dao.get_packages(channel.name, skip, limit, q, order_by)


@api_router.get(
    "/channels/{channel_name}/packages/{package_name}",
    response_model=rest_models.Package,
    tags=["packages"],
)
def get_package(package: db_models.Package = Depends(get_package_or_fail)):
    return package


@api_router.delete(
    "/channels/{channel_name}/packages/{package_name}",
    tags=["packages"],
)
def delete_package(
    background_tasks: BackgroundTasks,
    package: db_models.Package = Depends(get_package_or_fail),
    db=Depends(get_db),
    auth: authorization.Rules = Depends(get_rules),
    dao: Dao = Depends(get_dao),
):
    auth.assert_package_delete(package)

    filenames = [
        os.path.join(version.platform, version.filename)
        for version in package.package_versions  # type: ignore
    ]

    # get current platform containing the package
    platforms = set(
        [version.platform for version in package.package_versions]  # type: ignore
    )

    channel_name = package.channel_name

    db.delete(package)
    db.commit()

    for filename in filenames:
        pkgstore.delete_file(channel_name, filename)

    dao.update_channel_size(channel_name)

    wrapped_bg_task = background_task_wrapper(indexing.update_indexes, logger)
    # Background task to update indexes
    background_tasks.add_task(wrapped_bg_task, dao, pkgstore, channel_name, platforms)


@api_router.post(
    "/channels/{channel_name}/packages", status_code=201, tags=["packages"]
)
def post_package(
    new_package: rest_models.Package,
    channel: db_models.Channel = Depends(
        ChannelChecker(allow_proxy=False, allow_mirror=False),
    ),
    auth: authorization.Rules = Depends(get_rules),
    dao: Dao = Depends(get_dao),
):
    user_id = auth.assert_user()
    auth.assert_create_package(channel.name)
    pm.hook.validate_new_package(
        channel_name=channel.name,
        package_name=new_package.name,
        file_handler=None,
        condainfo=None,
    )
    package = dao.get_package(channel.name, new_package.name)
    if package:
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Package {channel.name}/{new_package.name} exists",
        )

    dao.create_package(channel.name, new_package, user_id, authorization.OWNER)


@api_router.get(
    "/channels/{channel_name}/members",
    response_model=List[rest_models.Member],
    tags=["channels"],
)
def get_channel_members(
    channel: db_models.Channel = Depends(get_channel_or_fail),
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    auth.assert_list_channel_members(channel.name)
    member_list = dao.get_channel_members(channel.name)

    return member_list


@api_router.post("/channels/{channel_name}/members", status_code=201, tags=["channels"])
def post_channel_member(
    new_member: rest_models.PostMember,
    channel: db_models.Channel = Depends(get_channel_or_fail),
    dao: Dao = Depends(get_dao),
    db=Depends(get_db),
    auth: authorization.Rules = Depends(get_rules),
):
    if not dao.get_user_by_username(new_member.username):
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"user {new_member.username} not found",
        )

    auth.assert_list_channel_members(channel.name)
    channel_member = dao.get_channel_member(channel.name, new_member.username)

    auth.assert_add_channel_member(channel.name, new_member.role)

    if channel_member:
        channel_member.role = new_member.role
        db.commit()
    else:
        dao.create_channel_member(channel.name, new_member)


@api_router.delete("/channels/{channel_name}/members", tags=["channels"])
def delete_channel_member(
    username: str,
    channel: db_models.Channel = Depends(get_channel_or_fail),
    dao: Dao = Depends(get_dao),
    db=Depends(get_db),
    auth: authorization.Rules = Depends(get_rules),
):
    if not dao.get_user_by_username(username):
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"user {username} not found",
        )

    auth.assert_list_channel_members(channel.name)
    channel_member = dao.get_channel_member(channel.name, username)

    if not channel_member:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"user {username} is not a member of channel {channel.name}",
        )

    auth.assert_remove_channel_member(channel.name, channel_member.role)

    db.delete(channel_member)
    db.commit()


@api_router.get(
    "/channels/{channel_name}/packages/{package_name}/members",
    response_model=List[rest_models.Member],
    tags=["packages"],
)
def get_package_members(
    package: db_models.Package = Depends(get_package_or_fail),
    dao: Dao = Depends(get_dao),
):
    member_list = dao.get_package_members(package.channel.name, package.name)

    return member_list


@api_router.post(
    "/channels/{channel_name}/packages/{package_name}/members",
    status_code=201,
    tags=["packages"],
)
def post_package_member(
    new_member: rest_models.PostMember,
    package: db_models.Package = Depends(get_package_or_fail),
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    auth.assert_add_package_member(package.channel.name, package.name, new_member.role)

    channel_member = dao.get_package_member(
        package.channel.name, package.name, new_member.username
    )
    if channel_member:
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=(
                f"Member {new_member.username} in "
                f"{package.channel.name}/{package.name} exists"
            ),
        )

    dao.create_package_member(package.channel.name, package.name, new_member)


@api_router.get(
    "/channels/{channel_name}/packages/{package_name}/versions",
    response_model=List[rest_models.PackageVersion],
    tags=["packages"],
)
def get_package_versions(
    package: db_models.Package = Depends(get_package_or_fail),
    dao: Dao = Depends(get_dao),
    time_created__ge: datetime.datetime = None,
    version_match_str: str = None,
):
    version_profile_list = dao.get_package_versions(
        package, time_created__ge, version_match_str
    )
    version_list = []

    for version, profile, api_key_profile in version_profile_list:
        version_data = rest_models.PackageVersion.from_orm(version)
        version_list.append(version_data)

    return version_list


@api_router.get(
    "/paginated/channels/{channel_name}/packages/{package_name}/versions",
    response_model=rest_models.PaginatedResponse[rest_models.PackageVersion],
    tags=["packages"],
)
def get_paginated_package_versions(
    package: db_models.Package = Depends(get_package_or_fail),
    dao: Dao = Depends(get_dao),
    skip: int = 0,
    limit: int = PAGINATION_LIMIT,
    time_created__ge: datetime.datetime = None,
    version_match_str: str = None,
):
    version_profile_list = dao.get_package_versions(
        package, time_created__ge, version_match_str, skip, limit
    )
    version_list = []

    for version, profile, api_key_profile in version_profile_list['result']:
        version_data = rest_models.PackageVersion.from_orm(version)
        version_list.append(version_data)

    return {
        'pagination': version_profile_list['pagination'],
        'result': version_list,
    }


@api_router.get(
    "/channels/{channel_name}/packages/{package_name}/versions/{platform}/{filename}",
    response_model=rest_models.PackageVersion,
    tags=["packages"],
)
def get_package_version(
    platform: str,
    filename: str,
    package_name: str,
    channel_name: str,
    package: db_models.Package = Depends(get_package_or_fail),
    dao: Dao = Depends(get_dao),
):
    version = dao.get_package_version_by_filename(
        channel_name, package_name, filename, platform
    )

    if not version:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"package version {platform}/{filename} not found",
        )

    return version


@api_router.delete(
    "/channels/{channel_name}/packages/{package_name}/versions/{platform}/{filename}",
    tags=["packages"],
)
def delete_package_version(
    platform: str,
    filename: str,
    channel_name: str,
    package_name: str,
    background_tasks: BackgroundTasks,
    dao: Dao = Depends(get_dao),
    db=Depends(get_db),
    auth: authorization.Rules = Depends(get_rules),
):
    version = dao.get_package_version_by_filename(
        channel_name, package_name, filename, platform
    )

    auth.assert_package_delete(version.package)

    if not version:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"package version {platform}/{filename} not found",
        )

    db.delete(version)
    db.commit()

    path = os.path.join(platform, filename)
    pkgstore.delete_file(channel_name, path)

    dao.cleanup_channel_db(channel_name, package_name)
    dao.update_channel_size(channel_name)

    wrapped_bg_task = background_task_wrapper(indexing.update_indexes, logger)
    # Background task to update indexes
    background_tasks.add_task(wrapped_bg_task, dao, pkgstore, channel_name, [platform])


@api_router.get(
    "/packages/search/", response_model=List[rest_models.PackageSearch], tags=["search"]
)
def search(
    q: str,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    user_id = auth.get_user()
    keywords, filters = parse_query('package', q)
    return dao.search_packages(keywords, filters, user_id)


@api_router.get(
    "/channels/search/", response_model=List[rest_models.ChannelSearch], tags=["search"]
)
def channel_search(
    q: str,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    user_id = auth.get_user()
    keywords, filters = parse_query('channel', q)
    return dao.search_channels(keywords, filters, user_id)


@api_router.get("/api-keys", response_model=List[rest_models.ApiKey], tags=["API keys"])
def get_api_keys(
    dao: Dao = Depends(get_dao), auth: authorization.Rules = Depends(get_rules)
):
    """Get API keys for current user"""

    user_id = auth.assert_user()
    user_role_keys, custom_role_keys = dao.get_api_keys_with_members(user_id)

    api_keys = []

    for key in user_role_keys:
        api_keys.append(
            rest_models.ApiKey(
                key=key.key,
                description=key.description,
                time_created=key.time_created,
                expire_at=key.expire_at,
                roles=None,
            )
        )

    from itertools import groupby

    grouped_by_key = groupby(custom_role_keys, key=lambda k: k[0])

    for group_key, group_items in grouped_by_key:
        roles = []
        for _, package_member, channel_member in group_items:
            if package_member:
                roles.append(
                    CPRole(
                        channel=package_member.channel_name,
                        package=package_member.package_name,
                        role=package_member.role,
                    )
                )
            if channel_member:
                roles.append(
                    CPRole(
                        channel=channel_member.channel_name,
                        package=None,
                        role=channel_member.role,
                    )
                )
        api_keys.append(
            rest_models.ApiKey(
                key=group_key.key,
                description=group_key.description,
                time_created=group_key.time_created,
                expire_at=group_key.expire_at,
                roles=roles,
            )
        )

    return api_keys


@api_router.post(
    "/api-keys", status_code=201, tags=["API keys"], response_model=rest_models.ApiKey
)
def post_api_key(
    api_key: rest_models.BaseApiKey,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    auth.assert_create_api_key_roles(api_key.roles)

    user_id = auth.assert_user()

    key = generate_random_key(32)
    dao.create_api_key(user_id, api_key, key)

    user_role_keys, custom_role_keys = dao.get_api_keys_with_members(user_id, key)

    if len(user_role_keys) > 0:
        key = user_role_keys[0]
        return rest_models.ApiKey(
            key=key.key,
            description=key.description,
            time_created=key.time_created,
            expire_at=key.expire_at,
            roles=None,
        )

    else:
        key = custom_role_keys[0][0]
        package_member = custom_role_keys[0][1]
        channel_member = custom_role_keys[0][2]
        roles = []
        if package_member:
            roles.append(
                CPRole(
                    channel=package_member.channel_name,
                    package=package_member.package_name,
                    role=package_member.role,
                )
            )

        if channel_member:
            roles.append(
                CPRole(
                    channel=channel_member.channel_name,
                    package=None,
                    role=channel_member.role,
                )
            )

        return rest_models.ApiKey(
            key=key.key,
            description=key.description,
            time_created=key.time_created,
            expire_at=key.expire_at,
            roles=roles,
        )


@api_router.delete("/api-keys/{key}", tags=["API keys"])
def delete_api_keys(
    key: str,
    dao: Dao = Depends(get_dao),
    db: Session = Depends(get_db),
    auth: authorization.Rules = Depends(get_rules),
):
    api_key = dao.get_api_key(key)

    if not api_key:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND, detail=f"key '{key}' does not exist"
        )

    auth.assert_delete_api_key(api_key)

    api_key.deleted = True

    db.commit()


@api_router.post(
    "/channels/{channel_name}/packages/{package_name}/files/",
    status_code=201,
    tags=["files"],
)
def post_file_to_package(
    background_tasks: BackgroundTasks,
    files: List[UploadFile] = File(...),
    force: Optional[bool] = Form(None),
    package: db_models.Package = Depends(get_package_or_fail),
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
    channel: db_models.Channel = Depends(
        ChannelChecker(allow_proxy=False, allow_mirror=False),
    ),
):
    handle_package_files(package.channel, files, dao, auth, force, package=package)
    dao.update_channel_size(package.channel_name)

    wrapped_bg_task = background_task_wrapper(indexing.update_indexes, logger)
    # Background task to update indexes
    background_tasks.add_task(wrapped_bg_task, dao, pkgstore, package.channel_name)


@api_router.post(
    "/channels/{channel_name}/upload/{filename}", status_code=201, tags=["upload"]
)
async def post_upload(
    background_tasks: BackgroundTasks,
    request: Request,
    channel_name: str,
    filename: str,
    sha256: str,
    force: bool = False,
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    logger.debug(
        f"Uploading file {filename} with checksum {sha256} to channel {channel_name}"
    )

    upload_hash = hashlib.sha256()

    body = TemporaryFile()
    async for chunk in request.stream():
        body.write(chunk)
        upload_hash.update(chunk)

    if sha256 and upload_hash.hexdigest() != sha256:
        raise HTTPException(
            status_code=status.HTTP_406_NOT_ACCEPTABLE, detail="Wrong SHA256 checksum"
        )

    user_id = auth.assert_user()
    auth.assert_create_package(channel_name)
    condainfo = CondaInfo((body), filename)
    dest = os.path.join(condainfo.info["subdir"], filename)

    body.seek(0)
    pkgstore.add_package(body, channel_name, dest)

    package_name = str(condainfo.info.get("name"))
    package_data = rest_models.Package(
        name=package_name,
        summary=str(condainfo.about.get("summary", "n/a")),
        description=str(condainfo.about.get("description", "n/a")),
    )
    if not dao.get_package(channel_name, package_name):
        dao.create_package(
            channel_name,
            package_data,
            user_id,
            authorization.OWNER,
        )

    # Update channeldata info
    dao.update_package_channeldata(channel_name, package_name, condainfo.channeldata)

    try:
        version = dao.create_version(
            channel_name=channel_name,
            package_name=package_name,
            package_format=condainfo.package_format,
            platform=condainfo.info["subdir"],
            version=condainfo.info["version"],
            build_number=condainfo.info["build_number"],
            build_string=condainfo.info["build"],
            size=condainfo.info["size"],
            filename=filename,
            info=json.dumps(condainfo.info),
            uploader_id=user_id,
            upsert=force,
        )
    except IntegrityError:
        logger.debug(f"duplicate package '{package_name}' in channel '{channel_name}'")
        raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail="Duplicate")

    pm.hook.post_add_package_version(version=version, condainfo=condainfo)

    wrapped_bg_task = background_task_wrapper(indexing.update_indexes, logger)
    # Background task to update indexes
    background_tasks.add_task(wrapped_bg_task, dao, pkgstore, channel_name)


@api_router.post("/channels/{channel_name}/files/", status_code=201, tags=["files"])
def post_file_to_channel(
    background_tasks: BackgroundTasks,
    files: List[UploadFile] = File(...),
    force: Optional[bool] = Form(None),
    channel: db_models.Channel = Depends(
        ChannelChecker(allow_proxy=False, allow_mirror=False)
    ),
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
):
    handle_package_files(channel, files, dao, auth, force)

    dao.update_channel_size(channel.name)

    wrapped_bg_task = background_task_wrapper(indexing.update_indexes, logger)
    # Background task to update indexes
    background_tasks.add_task(wrapped_bg_task, dao, pkgstore, channel.name)


@retry(
    stop=stop_after_attempt(3),
    retry=(retry_if_result(lambda x: x is None)),
    wait=wait_exponential(multiplier=1, min=4, max=10),
    after=after_log(logger, logging.WARNING),
)
def _extract_and_upload_package(file, channel_name, channel_proxylist):
    try:
        conda_info = CondaInfo(file.file, file.filename)
    except Exception as e:
        logger.error(
            f"Could not extract conda-info from package {file.filename}\n{str(e)}"
        )
        raise e

    dest = os.path.join(conda_info.info["subdir"], file.filename)
    parts = file.filename.rsplit("-", 2)

    if parts[0] != conda_info.info["name"]:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"package file name and info files do not match {file.filename}",
        )

    if channel_proxylist and conda_info.info["name"] in channel_proxylist:
        # do not upload files that are proxied
        logger.info(f"Skip upload of proxied file {file.filename}")
        return conda_info

    try:
        file.file.seek(0)
        logger.debug(
            f"Uploading file {dest} from channel {channel_name} to package store"
        )
        pkgstore.add_package(file.file, channel_name, dest)
    except AttributeError as e:
        logger.error(f"Could not upload {file}, {file.filename}. {str(e)}")
        return None

    return conda_info


def handle_package_files(
    channel,
    files,
    dao,
    auth,
    force,
    package=None,
    is_mirror_op=False,
):
    user_id = auth.assert_user()

    # quick fail if not allowed to upload
    # note: we're checking later that `parts[0] == conda_info.package_name`
    total_size = 0
    for file in files:
        parts = file.filename.rsplit("-", 2)
        if len(parts) != 3:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=f"package file name has wrong format {file.filename}",
            )
        else:
            package_name = parts[0]
        auth.assert_upload_file(channel.name, package_name)
        if force:
            auth.assert_overwrite_package_version(channel.name, package_name)

        # workaround for https://github.com/python/cpython/pull/3249
        if type(file.file) is SpooledTemporaryFile and not hasattr(file, "seekable"):
            file.file.seekable = file.file._file.seekable

        file.file.seek(0, os.SEEK_END)
        size = file.file.tell()
        total_size += size
        file.file.seek(0)

    dao.assert_size_limits(channel.name, total_size)

    channel_proxylist = []
    if channel.mirror_mode:
        if not is_mirror_op:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="Cannot upload packages to mirror channel",
            )
        else:
            channel_proxylist = json.loads(channel.channel_metadata).get(
                'proxylist', []
            )

    pkgstore.create_channel(channel.name)
    nthreads = config.general_package_unpack_threads
    with ThreadPoolExecutor(max_workers=nthreads) as executor:
        try:
            conda_infos = [
                ci
                for ci in executor.map(
                    _extract_and_upload_package,
                    files,
                    (channel.name,) * len(files),
                    (channel_proxylist,) * len(files),
                )
            ]
        except exceptions.PackageError as e:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST, detail=e.detail
            )

    conda_infos = [ci for ci in conda_infos if ci is not None]

    for file, condainfo in zip(files, conda_infos):
        logger.debug(f"Handling {condainfo.info['name']} -> {file.filename}")

        def _delete_file(condainfo, filename):
            dest = os.path.join(condainfo.info["subdir"], file.filename)
            pkgstore.delete_file(channel.name, dest)

        package_type = "tar.bz2" if file.filename.endswith(".tar.bz2") else "conda"
        UPLOAD_COUNT.labels(
            channel=channel.name,
            platform=condainfo.info["subdir"],
            package_name=condainfo.info["name"],
            version=condainfo.info["version"],
            package_type=package_type,
        ).inc()

        package_name = condainfo.info["name"]
        parts = file.filename.rsplit("-", 2)

        # check that the filename matches the package name
        # TODO also validate version and build string
        if parts[0] != condainfo.info["name"]:
            _delete_file(condainfo, file.filename)
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="filename does not match package name",
            )
        if package and (parts[0] != package.name or package_name != package.name):
            _delete_file(condainfo, file.filename)
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=(
                    f"requested package endpoint '{package.name}' "
                    f"does not match the uploaded package name '{parts[0]}'"
                ),
            )

        if not package and not dao.get_package(channel.name, package_name):
            try:
                if not channel_proxylist or package_name not in channel_proxylist:
                    pm.hook.validate_new_package(
                        channel_name=channel.name,
                        package_name=package_name,
                        file_handler=file.file,
                        condainfo=condainfo,
                    )
                    # validate uploaded package size and existence
                    try:
                        pkgsize, _, _ = pkgstore.get_filemetadata(
                            channel.name, f"{condainfo.info['subdir']}/{file.filename}"
                        )
                        if pkgsize != condainfo.info['size']:
                            raise errors.ValidationError(
                                f"Uploaded package {file.filename} "
                                "file size is wrong! Deleting"
                            )
                    except FileNotFoundError:
                        raise errors.ValidationError(
                            f"Uploaded package {file.filename} "
                            "file did not upload correctly!"
                        )

                package_data = rest_models.Package(
                    name=package_name,
                    summary=str(condainfo.about.get("summary", "n/a")),
                    description=str(condainfo.about.get("description", "n/a")),
                )
            except pydantic.main.ValidationError as err:
                _delete_file(condainfo, file.filename)
                raise errors.ValidationError(
                    "Validation Error for package: "
                    + f"{channel.name}/{file.filename}: {str(err)}"
                )
            except errors.ValidationError as err:
                _delete_file(condainfo, file.filename)
                logger.error(
                    f"Validation error in: {channel.name}/{file.filename}: {str(err)}"
                )
                raise err

            dao.create_package(
                channel.name,
                package_data,
                user_id,
                authorization.OWNER,
            )

        # Update channeldata info
        dao.update_package_channeldata(
            channel.name, package_name, condainfo.channeldata
        )

        try:
            version = dao.create_version(
                channel_name=channel.name,
                package_name=package_name,
                package_format=condainfo.package_format,
                platform=condainfo.info["subdir"],
                version=condainfo.info["version"],
                build_number=condainfo.info["build_number"],
                build_string=condainfo.info["build"],
                size=condainfo.info["size"],
                filename=file.filename,
                info=json.dumps(condainfo.info),
                uploader_id=user_id,
                upsert=force,
            )
        except IntegrityError:
            logger.error(
                f"duplicate package '{package_name}' in channel '{channel.name}'"
            )
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT, detail="Duplicate"
            )

        pm.hook.post_add_package_version(version=version, condainfo=condainfo)


app.include_router(
    api_router,
    prefix="/api",
)


@app.get("/api/.*", status_code=404, include_in_schema=False)
def invalid_api():
    return None


@app.on_event("startup")
def start_sync_download_counts():
    global download_counts
    wait_time = 1  # seconds

    db_manager = contextmanager(get_db)

    def commit_to_db(counts):
        count: int
        package = Tuple[str, str, str]

        with TicToc("sync download counts"):
            with db_manager(config) as db:
                dao = get_dao(db)
                for package, count in counts.items():
                    dao.incr_download_count(*package, incr=count)

    async def task():
        last_download_sync = None
        increment_delay = datetime.timedelta(seconds=DOWNLOAD_INCREMENT_DELAY_SECONDS)
        try:
            while True:
                n_total_downloads = sum(download_counts.values())
                now = datetime.datetime.utcnow()
                if n_total_downloads < DOWNLOAD_INCREMENT_MAX_DOWNLOADS:
                    if not n_total_downloads or (
                        last_download_sync
                        and (last_download_sync + increment_delay) > now
                    ):
                        await asyncio.sleep(wait_time)
                        continue
                logger.debug(
                    "Download counts: time since last sync %s, n/o downloads: %s",
                    now - (last_download_sync or now),
                    n_total_downloads,
                )
                new_items = download_counts.copy()
                download_counts.clear()
                await run_in_threadpool(commit_to_db, new_items)
                last_download_sync = now
        except asyncio.CancelledError:
            commit_to_db(download_counts)
            download_counts.clear()
            raise

    app.sync_download_task = asyncio.create_task(task())


@app.on_event("shutdown")
async def stop_sync_donwload_counts():
    app.sync_download_task.cancel()
    try:
        await app.sync_download_task
    except asyncio.CancelledError:
        pass


@app.head("/get/{channel_name}/{path:path}")
@app.get("/get/{channel_name}/{path:path}")
def serve_path(
    path,
    channel: db_models.Channel = Depends(get_channel_allow_proxy),
    accept_encoding: Optional[str] = Header(None),
    session=Depends(get_remote_session),
    dao: Dao = Depends(get_dao),
):
    chunk_size = 10_000

    is_package_request = path.endswith((".tar.bz2", ".conda"))

    package_name = None
    if is_package_request:
        try:
            platform, filename = os.path.split(path)
            package_name, version, hash_end = filename.rsplit('-', 2)
            package_type = "tar.bz2" if hash_end.endswith(".tar.bz2") else "conda"
            DOWNLOAD_COUNT.labels(
                channel=channel.name,
                platform=platform,
                package_name=package_name,
                version=version,
                package_type=package_type,
            ).inc()
            download_counts[(channel.name, filename, platform)] += 1
        except ValueError:
            pass

    if is_package_request and channel.mirror_channel_url:
        # if we exclude the package from syncing, redirect to original URL
        channel_proxylist = json.loads(channel.channel_metadata).get('proxylist', [])
        if channel_proxylist and package_name and package_name in channel_proxylist:
            return RedirectResponse(f"{channel.mirror_channel_url}/{path}")

    if channel.mirror_channel_url and channel.mirror_mode == "proxy":
        repository = RemoteRepository(channel.mirror_channel_url, session)
        if not pkgstore.file_exists(channel.name, path):
            download_remote_file(repository, pkgstore, channel.name, path)
        elif path.endswith(".json"):
            # repodata.json and current_repodata.json are cached locally
            # for channel.ttl seconds
            _, fmtime, _ = pkgstore.get_filemetadata(channel.name, path)
            if time.time() - fmtime >= channel.ttl:
                download_remote_file(repository, pkgstore, channel.name, path)

    if (
        is_package_request or pkgstore.kind == "LocalStore"
    ) and pkgstore.support_redirect:
        return RedirectResponse(pkgstore.url(channel.name, path))

    def iter_chunks(fid):
        while True:
            data = fid.read(chunk_size)
            if not data:
                break
            yield data

    if path == "" or path.endswith("/"):
        path += "index.html"
    package_content_iter = None

    headers = {}
    if accept_encoding and 'gzip' in accept_encoding and path.endswith('.json'):
        # return gzipped response
        try:
            package_content_iter = iter_chunks(
                pkgstore.serve_path(channel.name, path + '.gz')
            )
            path += '.gz'
            headers['Content-Encoding'] = 'gzip'
            headers['Content-Type'] = 'application/json'
        except FileNotFoundError:
            pass

    while not package_content_iter:
        try:
            package_content_iter = iter_chunks(pkgstore.serve_path(channel.name, path))

        except FileNotFoundError:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"{channel.name}/{path} not found",
            )
        except IsADirectoryError:
            path += "/index.html"

    fsize, fmtime, fetag = pkgstore.get_filemetadata(channel.name, path)
    headers.update(
        {
            'Cache-Control': f'max-age={channel.ttl}',
            'Content-Size': str(fsize),
            'Last-Modified': formatdate(fmtime, usegmt=True),
            'ETag': fetag,
        }
    )
    return StreamingResponse(package_content_iter, headers=headers)


@app.get("/get/{channel_name}")
def serve_channel_index(
    channel: db_models.Channel = Depends(get_channel_allow_proxy),
    accept_encoding: Optional[str] = Header(None),
    session=Depends(get_remote_session),
    dao: Dao = Depends(get_dao),
):
    return serve_path("index.html", channel, accept_encoding, session, dao)


frontend.register(app)



import uuid
from datetime import datetime, timedelta
from enum import Enum

import sqlalchemy as sa

from quetz.db_models import UUID, Base


class IntervalType(Enum):
    hour = "H"
    day = "D"
    month = "M"
    year = "Y"

    @property
    def timedelta(self):
        if self == IntervalType.hour:
            return timedelta(hours=1)
        if self == IntervalType.day:
            return timedelta(days=1)
        raise ValueError(f"can not create timedelta for interval '{self.name}'")


def round_timestamp(timestamp, period):
    """round timestamp to nearest period"""
    now_interval = timestamp.replace(minute=0, second=0, microsecond=0)
    if period in [IntervalType.day, IntervalType.month, IntervalType.year]:
        now_interval = now_interval.replace(hour=0)
    if period in [IntervalType.month, IntervalType.year]:
        now_interval = now_interval.replace(day=1)
    if period == IntervalType.year:
        now_interval = now_interval.replace(month=1)
    return now_interval


def next_timestamp(timestamp: datetime, interval: IntervalType):
    """next timestamp advanced by interval time"""
    if interval in [IntervalType.day, IntervalType.hour]:
        return timestamp + interval.timedelta
    if interval == IntervalType.month:
        if timestamp.month == 12:
            return timestamp.replace(year=timestamp.year + 1, month=1)
        else:
            return timestamp.replace(month=timestamp.month + 1)
    if interval == IntervalType.year:
        return timestamp.replace(year=timestamp.year + 1)
    raise ValueError(f"interval {interval.name} not supported")


class PackageVersionMetric(Base):
    __tablename__ = "aggregated_metrics"

    id = sa.Column(UUID, default=lambda: uuid.uuid4().bytes, primary_key=True)

    channel_name = sa.Column(sa.String)
    platform = sa.Column(sa.String)

    filename = sa.Column(sa.String)

    metric_name = sa.Column(sa.String(255), nullable=False)
    period = sa.Column(sa.Enum(IntervalType))
    count = sa.Column(sa.Integer, server_default=sa.text("0"), nullable=False)
    timestamp = sa.Column(sa.DateTime(), nullable=False)

    __table_args__ = (
        sa.Index(
            'package_version_metric_index',
            channel_name,
            platform,
            filename,
            metric_name,
            period,
            timestamp,
        ),
        sa.UniqueConstraint(
            channel_name,
            platform,
            filename,
            metric_name,
            period,
            timestamp,
            name="package_version_metric_constraint",
        ),
    )

    def __repr__(self):
        return (
            f"PackageVersionMetric(metric_name={self.metric_name}, "
            f"period={self.period.value}, "
            f"timestamp={self.timestamp},count={self.count})"
        )



"""Tests for the database models"""
# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.

import uuid

from quetz.db_models import User


def test_user(db):
    user = User(id=uuid.uuid4().bytes, username='paul')
    db.add(user)
    db.commit()

    assert len(db.query(User).all()) == 1

    found = User.find(db, 'paul')
    assert found.username == user.username
    found = User.find(db, 'dave')
    assert found is None

    db.delete(user)
    db.commit()



"""case insensitive channel names

Revision ID: a3ffa287d074
Revises: a80fb051a659
Create Date: 2020-12-07 14:10:27.374580

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'a3ffa287d074'
down_revision = 'a80fb051a659'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    if op.get_context().dialect.name == 'postgresql':
        op.execute(
            "CREATE COLLATION IF NOT EXISTS nocase (provider = icu, "
            "locale = 'und-u-ks-level2', "
            "deterministic = false)"
        )
    with op.batch_alter_table("channels") as batch_op:
        batch_op.alter_column("name", type_=sa.types.String(100, collation="nocase"))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("channels") as batch_op:
        batch_op.alter_column("name", type_=sa.types.String(None, collation=None))
    if op.get_context().dialect.name == 'postgresql':
        op.execute("DROP COLLATION nocase;")
    # ### end Alembic commands ###
